"""Added created_timestamp column to namespace_request

Revision ID: 771ef7c868de
Revises: 6fc1d0cf350c
Create Date: 2018-10-08 23:06:50.252729

"""

# revision identifiers, used by Alembic.
revision = '771ef7c868de'
down_revision = '6fc1d0cf350c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from time import time

def now_timestamp():
    return int(time() * 1000)

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('rb_namespace_request', sa.Column('created_timestamp', sa.BigInteger(), nullable=False, server_default=str(now_timestamp())))
    op.create_index(op.f('ix_rb_namespace_request_created_timestamp'), 'rb_namespace_request', ['created_timestamp'], unique=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('rb_namespace_request', 'created_timestamp')



from collections import (
    OrderedDict,
    deque,
)

from sqlalchemy import (
    BigInteger, Boolean, Column, Integer, Unicode, UnicodeText, or_, func
)

from sqlalchemy.orm import relationship, backref

import uuid

from slugify import slugify

from remarkbox.lib import (
    timestamp_to_datetime,
    timestamp_to_ago_string,
    timestamp_to_date_string,
)

from remarkbox.lib.render import markdown_to_html

from .meta import Base, RBase
from .meta import UUIDType
from .meta import now_timestamp, foreign_key, get_object_by_id

from .uri import get_or_create_uri

from .event import NodeEvent

import codecs

import logging

try:
    unicode("")
except:
    from six import u as unicode

log = logging.getLogger(__name__)


class Node(RBase, Base):
    """
    Node class acts like a linked list to support nested:

        Question -> Answers -> Comments

        Thread -> Posts -> Remarks

    The `parent_id` column is only used by children nodes.

    The class will allow 3 layer nesting like StackOverflow or unlimited
    nesting like reddit.  The frontend code determines the particular
    nesting strategy.

    For more information checkout this section of the SQLAlchemy Docs:

    http://docs.sqlalchemy.org/en/latest/orm/relationships.html#adjacency-list-relationships
    """

    id = Column(UUIDType, primary_key=True, index=True)
    root_id = Column(UUIDType, index=True, default=None)
    parent_id = Column(UUIDType, foreign_key("Node", "id"), default=None)
    namespace_id = Column(UUIDType, foreign_key("Namespace", "id"), default=None)
    user_id = Column(UUIDType, foreign_key("User", "id"), default=None)
    user_surrogate_id = Column(
        UUIDType, foreign_key("UserSurrogate", "id"), default=None
    )
    title = Column(Unicode(256), default=None)
    data = Column(UnicodeText, default=None)
    data_html = Column(UnicodeText, default=None)
    # the depth of this node in the thread's graph / tree.
    graph_depth = Column(Integer, nullable=False, default=-1)
    # TODO: someday this should be renamed to created_timestamp
    created = Column(BigInteger, nullable=False)
    # TODO: someday this should be renamed to changed_timestamp
    changed = Column(BigInteger, nullable=False)
    disabled_timestamp = Column(BigInteger)
    disabled = Column(Boolean, default=False)
    verified = Column(Boolean, default=False)
    # a locked thread (root node) prevents new commenting.
    locked = Column(Boolean, default=False)
    # by default comments are approved. unless Namespace hide_unless_approved.
    approved = Column(Boolean, default=True)
    # is there a related Uri model to this node?
    has_uri = Column(Boolean, default=False, nullable=False)
    ip_address = Column(Unicode(45), default=None)

    # lazy='joined' performs a left join to reduce queries, it's magic.
    user = relationship(argument="User", lazy="joined")

    # lazy='joined' performs a left join to reduce queries, it's magic.
    user_surrogate = relationship(argument="UserSurrogate", lazy="joined")

    # lazy='dynamic' returns a query object instead of collection.
    # Reference: http://docs.sqlalchemy.org/en/latest/orm/collections.html
    children = relationship(
        argument="Node",
        lazy="dynamic",
        order_by="Node.created",
        foreign_keys=[parent_id],
        backref=backref("parent", remote_side=[id]),
    )

    # lazy='dynamic' returns a query object instead of collection.
    # Reference: http://docs.sqlalchemy.org/en/latest/orm/collections.html
    children_desc = relationship(
        argument="Node",
        lazy="dynamic",
        order_by="desc(Node.created)",
        foreign_keys=[parent_id],
    )

    votes = relationship(argument="Vote", backref="node", order_by="desc(Vote.created)")

    # set uselist to false to create one-to-one relationship.
    uri = relationship(argument="Uri", uselist=False, back_populates="node")

    # set uselist to false to create one-to-one relationship.
    namespace = relationship(argument="Namespace", uselist=False)

    # lazy='joined' performs a left join to reduce queries, it's magic.
    # this is used for denormalized stats like counts, etc.
    cache = relationship(
        argument="NodeCache", lazy="joined", back_populates="node", uselist=False
    )

    watchers = relationship(argument="Watcher", lazy="dynamic", back_populates="node")

    events = relationship(argument="NodeEvent", lazy="dynamic", back_populates="node")

    def __init__(self):
        self.id = uuid.uuid1()
        self.created = now_timestamp()
        self.changed = self.created

    def get_children(self, order_by="asc"):
        if order_by == "desc":
            return self.children_desc
        return self.children

    @property
    def verified_children(self):
        return self.children.filter(Node.verified == True)

    @property
    def unverified_children(self):
        return self.children.filter(Node.verified == False)

    @property
    def path_to_root(self):
        """The path from this node to the root node."""
        # TODO: memoize this result? how / when to deal with cache busting?
        node = self
        path_to_root = []
        path_to_root.append(node)
        while node.parent is not None:
            node = node.parent
            path_to_root.append(node)
        return path_to_root

    @property
    def path_from_root(self):
        """The path from the root node to this node."""
        return list(reversed(self.path_to_root))

    @property
    def root2(self):
        """The origin or top most parent node without a parent."""
        if self.is_root:
            return self
        return self.path_to_root[-1]

    @property
    def root(self):
        """The origin or top most parent node without a parent."""
        if self.is_root:
            return self
        return self.dbsession.query(Node).filter(Node.id == self.root_id).one()

    @property
    def is_root(self):
        if not self.parent_id or not self.root_id or self.root_id == self.id:
            return True
        return False

    def recompute_depth(self):
        """recompute the depth of a node in the thread."""
        self.graph_depth = len(self.path_to_root) - 1

    @property
    def depth(self):
        if self.graph_depth == -1:
            self.recompute_depth()
        return self.graph_depth

    @property
    def stats(self):
        """The denormalized stats for this node."""
        from .node_cache import NodeCache

        if self.cache is None:
            self.cache = NodeCache(self.id)
        return self.cache.stats

    def score(self):
        # this doesn't actually exist right now. I removed voting from this
        # iteration of the codebase. maybe worth adding in at some point.
        return sum([vote.value for vote in self.votes])

    def children_order_by_score(self):
        # TODO: this should be optimized, we should craft a query and have
        # database to return results in the expected order instead of multiple
        # queries.  First we select children then we sort by selecting each
        # vote for each child and sum the result.
        return sorted(self.children, key=lambda x: x.score(), reverse=True)

    def new_child(self):
        """create and return new child node of this node, in memory"""
        child = Node()
        child.parent = self
        child.root_id = self.root_id
        return child

    def new_event(self, user, action):
        """create and return new node_event for this node, in memory"""
        event = NodeEvent(node=self, user=user, action=action)
        self.events.append(event)
        return event

    def avatar_uri(self, **kwargs):
        """Return invatar URI. Optionally return gravatar URI."""
        u = self.user if self.user else self.user_surrogate
        if self.disabled:
            kwargs["text"] = "-"
            kwargs["bg"] = "#aaaaaa"
        if self.disabled or not self.verified:
            return u.svgtar(**kwargs)
        return u.avatar_uri(**kwargs)

    @property
    def created_timestamp(self):
        return self.created

    @property
    def human_created_timestamp(self):
        return timestamp_to_ago_string(self.created)

    # TODO: take into account user's timezone?
    @property
    def created_date(self):
        return timestamp_to_date_string(self.created)

    # TODO: take into account user's timezone?
    @property
    def created_datetime(self):
        return timestamp_to_datetime(self.created)

    @property
    def changed_timestamp(self):
        return self.changed

    @property
    def human_changed_timestamp(self):
        return timestamp_to_ago_string(self.changed)

    # TODO: take into account user's timezone?
    @property
    def changed_date(self):
        return timestamp_to_date_string(self.changed)

    # TODO: take into account user's timezone?
    @property
    def changed_datetime(self):
        return timestamp_to_datetime(self.changed)

    @property
    def title_ascii(self):
        """return the ascii version of the title, ignoring errors or None."""
        if self.title:
            return self.title.encode('ascii', 'ignore')

    @property
    def slug(self):
        """return slug from title or empty string"""
        if self.title:
            return slugify(self.title)
        return ""

    @property
    def path(self):
        """return a node's path part of uri"""
        if self.slug:
            return "/{}/{}".format(self.id, self.slug)
        else:
            return "/{}".format(self.id)

    @property
    def short_id(self):
        """return the base64 representation of UUID: 36 -> 22 unicode chars"""
        base_64_id = codecs.encode(self.id.bytes, "base64")
        base_64_id = base_64_id.rstrip(b"=\n").replace(b"/", b"_").replace(b"+", b"-")
        return base_64_id.decode("utf-8")

    @property
    def was_edited(self):
        return not self.created == self.changed

    @property
    def enabled(self):
        return not self.disabled

    def set_data(self, data, namespace=None):
        if namespace is None:
            namespace = self.root.namespace
        self.data = data
        self.data_html = markdown_to_html(data, namespace)

    def _invalidate_cache(self):
        if self.root.cache:
            self.root.cache.invalidate()

    def edit(self, data):
        """edit node."""
        self.set_data(data)
        self.changed = now_timestamp()
        self._invalidate_cache()

    def disable(self):
        """disable node."""
        self.disabled = True
        self.disabled_timestamp = now_timestamp()
        self._invalidate_cache()

    def enable(self):
        """enable node."""
        self.disabled = False
        self.disabled_timestamp = None
        self._invalidate_cache()

    def verify(self):
        """verify node."""
        self.verified = True
        self._invalidate_cache()

    def unverify(self):
        """unverify node."""
        self.verified = False
        self._invalidate_cache()

    def approve(self):
        """verify node."""
        self.approved = True
        self._invalidate_cache()

    def deny(self):
        """unverify node."""
        self.approved = False
        self._invalidate_cache()

    def visible(self):
        if self.disabled:
            return False
        return True

    def is_owner(self, user):
        """Return True if given user owns this node, else False."""
        if user and user.authenticated and user == self.user:
            return True
        return False


def get_node_by_id(dbsession, node_id):
    """Try to get Node object by id or return None."""
    return get_object_by_id(dbsession, node_id, Node)


def get_node_by_uri(dbsession, node_uri):
    return get_or_create_uri(dbsession, node_uri).node


def create_root_node():
    node = Node()
    node.root_id = node.id
    node.graph_depth = 0
    return node


def get_or_create_node_by_uri(dbsession, node_uri, node_title=None):
    """Try to get Node object by uri or return new Node"""
    uri = get_or_create_uri(dbsession, node_uri)

    if uri.node is None:
        # create new root node for external page (embed).

        # imported here to prevent circular import.
        from .namespace import get_or_create_namespace

        uri.node = create_root_node()
        uri.node.verified = True
        uri.node.has_uri = True
        uri.node.title = node_title
        uri.node.namespace = get_or_create_namespace(dbsession, uri.parsed.hostname)

    elif node_title:
        # if the external page title has changed, update our database.
        if node_title != uri.node.title:
            uri.node.title = node_title
            dbsession.add(uri.node)
            dbsession.flush()

    return uri.node


def get_topsecret_roots(dbsession):
    """Return all verified parent nodes by descending created order."""
    return (
        dbsession.query(Node)
        .filter(Node.parent_id == None, Node.verified == True, Node.disabled == False)
        .order_by(Node.changed.desc())
    )


def get_topsecret_nodes(dbsession, limit=100, offset=0):
    """Return all nodes by descending created order."""
    return (
        dbsession.query(Node)
        .filter(Node.disabled == False, Node.user_id != None)
        .order_by(Node.changed.desc())
        .limit(limit)
        .offset(offset)
    )


def get_root_nodes(dbsession):
    """Return a list of all root nodes (nodes without a parent.)"""
    return (
        dbsession.query(Node)
        .filter(Node.parent_id == None)
        .order_by(Node.changed.desc())
    )


def get_root_nodes_by_keywords(dbsession, keywords, namespace=None):

    # make dictionary mapping of root_node_ids to accumulated scores
    root_scores = {}

    # make dictionary mapping of root_node_ids to root_node_objects
    root_nodes = {}

    # a list of matched node objects.
    nodes = []

    if namespace:
        node_query = namespace.visible_nodes
    else:
        node_query = get_nodes_who_share_roots_query(
            dbsession, get_topsecret_roots(dbsession)
        )

    for keyword in keywords:

        # extend nodes, with a list of nodes which match this keyword. 
        keyword_filter = Node.data.ilike("%{}%".format(keyword))
        nodes.extend(node_query.filter(keyword_filter).all())

    # accumulate scores and root node objects.
    for node in nodes:
        if node.root.id not in root_scores:
            root_scores[node.root.id] = 0
        root_scores[node.root.id] += 1

        if node.root.id not in root_nodes:
            root_nodes[node.root.id] = node.root

    # sort the scores dictionary by value (score),
    # but get the key (object id),
    # reversed (highest score first)
    sorted_root_ids = sorted(root_scores, key=root_scores.get, reverse=True)

    # return root node objects in sorted order by score.
    return [root_nodes[root_id] for root_id in sorted_root_ids]


def get_nodes_who_share_roots_query(dbsession, root_nodes):
    # a giant n "or" query, where n is the number of root_nodes.
    if root_nodes.count() == 0:
        # return a "null" query object. still hits database...
        # this allows us to pass around the query object.
        from sqlalchemy.sql import text
        return dbsession.query(Node).filter(text("0 == 1"))
    return dbsession.query(Node).filter(or_(Node.root_id == n.id for n in root_nodes))


def get_nodes_who_share_roots(dbsession, root_nodes):
    return get_nodes_who_share_roots_query(dbsession, root_nodes).all()


def get_nodes_who_share_root(dbsession, root_node, order="oldest-first"):
    nodes = dbsession.query(Node).filter(Node.root_id == root_node.id)
    if order == "oldest-first":
        nodes = nodes.order_by(Node.created)
    elif order == "newest-first":
        nodes = nodes.order_by(Node.created.desc())
    return nodes


def get_node_count_who_share_root(dbsession, root_node):
    return (
        dbsession.query(func.count("id"))
        .select_from(Node)
        .filter(Node.root_id == root_node.id)
        .scalar()
    )


def get_visible_node_count_who_share_root(dbsession, root_node):
    return (
        dbsession.query(func.count("id"))
        .select_from(Node)
        .filter(
            Node.root_id == root_node.id, Node.disabled == False, Node.approved == True
        )
        .scalar()
    )


def get_node_id_map(nodes):
    """return a dictionary of all nodes who share this root."""
    node_id_map = {}
    for node in nodes:
        node_id_map[node.id] = node
    return node_id_map


def get_graph_from_nodes(nodes):
    """infinite nested graph or tree of nodes."""
    graph = OrderedDict()

    # create all node keys in graph.
    for node in nodes:
        graph[node.id] = []

    for node in nodes:
        if node.parent is not None:
            graph[node.parent.id].append(node.id)

    return graph


def get_group_conversation_ids(nodes):
    """Return a list of node ids whose parent is root."""
    if len(list(nodes)) == 0:
        return []
    root = nodes[0].root
    return [node.id for node in nodes if node.parent and node.parent.id == root.id]


# Reference post:
# http://kmkeen.com/python-trees/2010-09-18-08-55-50-039.html
# strategy queue.
def flatten_graph(node_id, graph, include_given_node_id=False):
    """
    Given a node_id and graph (or tree),
    return a list of every descendant node_id, depth first order.
    """
    flat_graph = []

    to_crawl = deque([node_id])

    while to_crawl:
        current_id = to_crawl.pop()

        flat_graph.append(current_id)

        children_ids = graph[current_id]

        # reverse so that we extend the queue in the expected order.
        children_ids.reverse()

        to_crawl.extend(children_ids)

    if include_given_node_id:
        return flat_graph
    return flat_graph[1:]


def get_conversation_graph_from_nodes(nodes, graph=None, node_id_map=None, order="oldest-first"):
    """
    Given a list of nodes, return a grouped conversation graph which is a
    dictionary of ids where the keys are the top level conversation node ids
    and the values are a list of all decendant node ids.

    If you already have it computed, you may optionally provide an existing
    graph (dictionary tree of node ids), and/or an existing node_id_map.
    """
    def get_timestamp(node_id):
        return node_id_map[node_id].created

    if graph is None:
        graph = get_graph_from_nodes(nodes)

    if node_id_map is None:
        node_id_map = get_node_id_map(nodes)

    _reverse = False
    if order == "newest-first":
        _reverse = True

    convo_graph = OrderedDict()

    convo_ids = sorted(
        get_group_conversation_ids(nodes),
        key=get_timestamp,
        reverse=_reverse
    )

    for convo_node_id in convo_ids:
        descendant_ids = flatten_graph(
            convo_node_id,
            graph,
        )
        # sort by timestamp, oldest first.
        convo_graph[convo_node_id] = sorted(
            descendant_ids,
            key=get_timestamp,
        )

    return convo_graph


def recompute_depth_on_all_node_objects(dbsession, chunk_lenth=500):
    """Only use this if a many nodes were moved and/or re-nested."""

    #
    # WARNING, this could take a long time and be computationally expensive
    #          both on the host running it, and the database, respectively.
    #

    import math

    log.info("loading and querying nodes...")

    node_count = dbsession.query(Node).count()

    nodes = dbsession.query(Node).all()

    log.info("node_count={}".format(node_count))

    for i, node in enumerate(nodes):

        # this is the computationally intense part.
        node.recompute_depth()
        dbsession.add(node)

        if i % chunk_lenth == 0:
            progress = int(math.floor((i / float(node_count)) * 100))
            log.info(
                "Flushing chunk={}/{}, chunk_lenth={} progress={}%".format(
                    i, node_count, chunk_lenth, progress
                )
            )
            dbsession.flush()

    # one final flush to database for the last partial chunk.
    log.info(
        "Flushing chunk={}/{}, chunk_lenth={} progress={}%".format(
            i + 1, node_count, chunk_lenth, 100
        )
    )
    dbsession.flush()



from sqlalchemy import BigInteger, Boolean, Column, Enum

from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.associationproxy import association_proxy

from .meta import Base, RBase
from .meta import now_timestamp, foreign_key
from .meta import UUIDType


class PayWhatYouCan(RBase, Base):
    user_id = Column(UUIDType, foreign_key("User", "id"), primary_key=True, index=True)
    # whether or not this is a one time charge or a yearly.
    frequency = Column(Enum("once","yearly", name="frequency"), nullable=False)
    # the amount in cents to charge this user.
    amount = Column(BigInteger, nullable=False)
    # the total contributions from this user in cents.
    contributions = Column(BigInteger, default=0, nullable=False)
    created_timestamp = Column(BigInteger, nullable=False)
    updated_timestamp = Column(BigInteger, nullable=False)

    # 1-to-1 relationships.
    user = relationship(argument="User", uselist=False, lazy="joined", back_populates="pay_what_you_can")

    def __init__(self, user, frequency, amount):
        self.created_timestamp = now_timestamp()
        self.updated_timestamp = now_timestamp()
        self.user = user
        self.frequency = frequency
        self.amount = amount

    def update(self, frequency, amount):
        if frequency != self.frequency:
            self.frequency = frequency
            self.updated_timestamp = now_timestamp()
        if amount != self.amount:
            self.amount = amount
            self.updated_timestamp = now_timestamp()



# quote email address in OTP so that a plus address
# is not  decoded as a space during authentication.
try:
    # Python 2.
    from urllib import quote_plus
except ImportError:
    # Python 3.
    from urllib.parse import quote_plus

from remarkbox.lib.mail_messages import (
    WELCOME_1_TEXT,
    WELCOME_1_HTML,
    WELCOME_2_TEXT,
    WELCOME_2_HTML,
    OPERATOR_HTML,
)

import dkim

import smtplib

from email.mime.multipart import MIMEMultipart

from email.mime.text import MIMEText

# catch socket errors when postfix isn't running...
from socket import error as socket_error

import logging

log = logging.getLogger(__name__)

from jinja2 import Environment, PackageLoader, select_autoescape

jinja2_env = Environment(
    loader=PackageLoader("remarkbox", "templates"),
    autoescape=select_autoescape(["html", "xml"]),
)


def send_email(
    to_email,
    sender_email,
    subject,
    message_text,
    message_html,
    relay="localhost",
    dkim_private_key_path="",
    dkim_selector="",
    dkim_signature_algorithm="ed25519-sha256",
    debug_mode=False,
):
    # the `email` library assumes it is working with string objects.
    # the `dkim` library assumes it is working with byte objects.
    # this function performs the acrobatics to make them both happy.
    if isinstance(message_text, bytes):
        # needed for Python 3.
        message_text = message_text.decode()

    if isinstance(message_html, bytes):
        # needed for Python 3.
        message_html = message_html.decode()

    sender_domain = sender_email.split("@")[-1]
    msg = MIMEMultipart("alternative")
    msg.attach(MIMEText(message_text, "plain"))
    msg.attach(MIMEText(message_html, "html"))
    msg["To"] = to_email
    msg["From"] = sender_email
    msg["Subject"] = subject

    try:
        # Python 3 libraries expect bytes.
        msg_data = msg.as_bytes()
    except:
        # Python 2 libraries expect strings.
        msg_data = msg.as_string()

    if dkim_private_key_path and dkim_selector:
        try:
            # the dkim library uses regex on byte strings so everything
            # needs to be encoded from strings to bytes.
            with open(dkim_private_key_path) as fh:
                dkim_private_key = fh.read()
            headers = [b"To", b"From", b"Subject"]
            sig = dkim.sign(
                message=msg_data,
                selector=str(dkim_selector).encode(),
                domain=sender_domain.encode(),
                privkey=dkim_private_key.encode(),
                include_headers=headers,
                signature_algorithm=dkim_signature_algorithm.encode(),
            )
            # add the dkim signature to the email message headers.
            # decode the signature back to string_type because later on
            # the call to msg.as_string() performs it's own bytes encoding...
            msg["DKIM-Signature"] = sig[len("DKIM-Signature: ") :].decode()

            try:
                # Python 3 libraries expect bytes.
                msg_data = msg.as_bytes()
            except:
                # Python 2 libraries expect strings.
                msg_data = msg.as_string()
        except Exception as e:
            if debug_mode:
                log.error(f"DKIM signing failed: {str(e)}")
            raise

    try:
        s = smtplib.SMTP(relay)
        s.sendmail(sender_email, [to_email], msg_data)
        s.quit()
        return msg

    except (socket_error, smtplib.SMTPException) as e:
        error_msg = f"Failed to send email: {str(e)}"

        if debug_mode:
            # Log the error first for quick scanning
            log.error(error_msg)
            # Then log the email details
            log.info(
                f"""

Email Contents:
To: {to_email}
From: {sender_email}
Subject: {subject}

Text Content:
{message_text}

HTML Content:
{message_html}
            """
            )

        if not debug_mode:
            raise
        return None


def send_pyramid_email(request, to_email, subject, message_text, message_html):
    """Thin wrapper around `send_email` to customise settings using request object."""
    default_sender = "no-reply@{}".format(request.domain)
    sender_email = request.app.get("email.sender", default_sender)
    subject = "{} | {}".format(
        subject, request.app.get("email.subject_postfix", request.domain)
    )
    relay = request.app.get("email.relay", "localhost")
    dkim_private_key_path = request.app.get("email.dkim_private_key_path", "")
    dkim_selector = request.app.get("email.dkim_selector", "")
    dkim_signature_algorithm = request.app.get(
        "email.dkim_signature_algorithm", "ed25519-sha256"
    )

    send_email(
        to_email,
        sender_email,
        subject,
        message_text,
        message_html,
        relay,
        dkim_private_key_path,
        dkim_selector,
        dkim_signature_algorithm,
        request.debug_mode,
    )


def send_verification_digits_to_email(request, to_email, raw_digits):
    """
    Send email with raw_digits a user may pass to verify & authenticate.

    request
      the request (of the successful log in attempt)

    to_email
      the email address to send the raw_digits

    raw_digits:
      the raw (unencrypted) digits the user may use to verify & authenticate.
    """
    subject = "Verification Code - {}".format(raw_digits)

    message_text = WELCOME_1_TEXT.format(raw_digits)
    message_html = WELCOME_1_HTML.format(subject, raw_digits)

    if not request.user.verified:
        message_text = WELCOME_1_TEXT.format(raw_digits)
        message_html = WELCOME_1_HTML.format(subject, raw_digits)
    else:
        message_text = WELCOME_2_TEXT.format(raw_digits)
        message_html = WELCOME_2_HTML.format(subject, raw_digits)

    send_pyramid_email(request, to_email, subject, message_text, message_html)


def send_operator_email(request, msg):
    send_pyramid_email(
        request,
        "russell.ballestrini@gmail.com",
        msg,
        msg,
        OPERATOR_HTML.format(msg),
    )


def send_template_email(
    request, to_email, subject, text_template_name, html_template_name, context
):
    text_template = jinja2_env.get_template(text_template_name)
    html_template = jinja2_env.get_template(html_template_name)
    send_pyramid_email(
        request,
        to_email,
        subject,
        text_template.render(**context),
        html_template.render(**context),
    )



"""add hide_unless_approved column

Revision ID: 7a3a4755b4a8
Revises: 1040ee4836af
Create Date: 2017-08-09 21:53:01.995660

"""

# revision identifiers, used by Alembic.
revision = "7a3a4755b4a8"
down_revision = "1040ee4836af"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "Namespace", sa.Column("hide_unless_approved", sa.Boolean(), nullable=True)
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("Namespace", "hide_unless_approved")
    ### end Alembic commands ###



from sqlalchemy import Integer, BigInteger, Column

import uuid

from remarkbox.lib import timestamp_to_ago_string, timestamp_to_date_string

from .meta import Base, RBase
from .meta import now_timestamp, foreign_key
from .meta import UUIDType


class Vote(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    user_id = Column(UUIDType, foreign_key("User", "id"))
    node_id = Column(UUIDType, foreign_key("Node", "id"))
    created = Column(BigInteger, nullable=False)
    value = Column(Integer, nullable=False)

    def __init__(self):
        self.id = uuid.uuid1()
        self.created = now_timestamp()

    @property
    def human_timestamp(self):
        return timestamp_to_ago_string(self.created)

    # TODO: take into account user's timezone?
    @property
    def created_date(self):
        return timestamp_to_date_string(self.created)



from pyramid.view import view_config

from remarkbox.models import (
    get_topsecret_namespace_requests,
    get_topsecret_notifications,
    get_topsecret_namespaces,
    get_topsecret_roots,
    get_topsecret_nodes,
)

from remarkbox.views import super_fly_required


@view_config(route_name="topsecret-namespaces", renderer="list-namespaces.j2")
@super_fly_required
def topsecret_namespaces(request):
    return {
        "namespaces": get_topsecret_namespaces(request.dbsession),
        "the_title": "topsecret namespaces!",
    }


@view_config(
    route_name="topsecret-namespace-requests", renderer="list-namespace-requests.j2"
)
@super_fly_required
def topsecret_namespace_requests(request):
    limit = request.params.get("limit", 100)
    page = request.params.get("page", 0)
    return {
        "namespace_requests": get_topsecret_namespace_requests(
            request.dbsession,
            limit=limit,
            offset=page,
        ),
        "the_title": "topsecret namespace requests!",
    }


@view_config(
    route_name="topsecret-namespace-owners", renderer="list-namespace-owners.j2"
)
@super_fly_required
def topsecret_namespace_owners(request):
    return {
        "namespaces": get_topsecret_namespaces(request.dbsession),
        "the_title": "topsecret namespace owners!",
    }


@view_config(route_name="topsecret-notifications", renderer="list-notifications.j2")
@super_fly_required
def topsecret_notifications(request):
    return {
        "notifications": get_topsecret_notifications(
            request.dbsession, limit=request.page_size, offset=request.page_offset
        ),
        "the_title": "topsecret notifications!",
        "topsecret": True,
    }


@view_config(route_name="topsecret", renderer="home.j2")
@super_fly_required
def topsecret_roots(request):
    return {
        "nodes": get_topsecret_roots(request.dbsession),
        "the_title": "topsecret root nodes!",
    }


@view_config(route_name="topsecret-nodes", renderer="list-nodes.j2")
@super_fly_required
def topsecret_nodes(request):
    return {
        "nodes": get_topsecret_nodes(
            request.dbsession, limit=request.page_size, offset=request.page_offset
        ),
        "the_title": "topsecret activity on everything!",
    }



from sqlalchemy import (
    Integer,
    BigInteger,
    Boolean,
    Column,
    Enum,
    Unicode,
    UnicodeText,
    or_,
)

from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.associationproxy import association_proxy

import uuid

from collections import defaultdict

from .meta import Base, RBase
from .meta import UUIDType
from .meta import SUBSCRIPTION_TYPES
from .meta import now_timestamp, foreign_key

from .node import Node, get_nodes_who_share_roots_query

from .namespace_user import NamespaceUser

from .oauth import Oauth

from remarkbox.lib import timestamp_to_ago_string, timestamp_to_datetime

try:
    unicode("")
except:
    from six import u as unicode


# these attributes are protected such that they return the defaults
# when a Namespace is frozen or subscription is expired.
PROTECTED_ATTRIBUTES = {
    "description": None,
    "placeholder_text": None,
    "no_nodes_text": None,
    "theme": None,
    "theme_embed": None,
    "stylesheet": None,
    "stylesheet_uri": None,
    "stylesheet_embed": None,
    "stylesheet_embed_uri": None,
    "wiki": False,
    "google_analytics_id": None,
    "hide_unverified": False,
    "hide_unless_approved": False,
    "hide_powered_by": False,
    "mathjax": False,
    "link_protection": False,
    "ignore_query_string": False,
    "reverse_order": False,
    "group_conversations": False,
}


class Namespace(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    name = Column(Unicode(256), nullable=False, unique=True, index=True)
    description = Column(Unicode(256), default=None)
    # optional placeholder text for new node textarea.
    placeholder_text = Column(Unicode(256), default=None, nullable=True)
    # optional placeholder text when there are no nodes in thread.
    no_nodes_text = Column(Unicode(256), default=None, nullable=True)
    # the basic mode theme. TODO: change this column name to `theme_basic`.
    theme = Column(Unicode(256), default=None)
    # the embed mode theme.
    theme_embed = Column(Unicode(256), default=None)
    avatar_size = Column(Integer, default=None)
    # TODO: change this to `stylesheet_basic` and `stylesheet_basic_timestamp`.
    stylesheet = Column(UnicodeText, default=None)
    stylesheet_timestamp = Column(BigInteger)
    stylesheet_uri = Column(Unicode(256), default=None)
    stylesheet_embed = Column(UnicodeText, default=None)
    stylesheet_embed_timestamp = Column(BigInteger)
    stylesheet_embed_uri = Column(Unicode(256), default=None)
    # when was the last time we tested/verified the namespace owner request?
    owner_request_timestamp = Column(BigInteger, nullable=True)
    # optional google analytics id for stand alone mode.
    google_analytics_id = Column(Unicode(18), nullable=True)
    # allow anyone to edit the root node. (not implemented yet)
    wiki = Column(Boolean, default=False)
    # by default we show all nodes but a Namespace can change this behavior.
    # should a node be hidden if unverified by owner or moderator?
    hide_unverified = Column(Boolean, default=False)
    # should a node be hidden until approved by a moderator?
    hide_unless_approved = Column(Boolean, default=False)
    # should we hide the poweredby Remarkbox logo?
    hide_powered_by = Column(Boolean, default=False)
    # should the list of root nodes in this namespace be public or hidden?
    public = Column(Boolean, default=False)
    # should we enable MathJax?
    mathjax = Column(Boolean, default=False)
    # should we enable Link Protection to prevent comments from having links?
    link_protection = Column(Boolean, default=False)
    # check this box if the external site does _not_ use
    # query strings for page identification or navigation.
    ignore_query_string = Column(Boolean, default=False)
    # should we reverse the node order and put newest on top?
    reverse_order = Column(Boolean, default=False)
    # should we group conversations and limit to nesting 2 deep?
    group_conversations = Column(Boolean, default=False)
    # the type of subscription of this Namespace.
    subscription_type = Column(
        Enum(*SUBSCRIPTION_TYPES, name="subscription_type"),
        # Hooray, Remarkbox becomes pay-what-you-can!
        # https://en.wikipedia.org/wiki/Pay_what_you_can
        default="production",
        nullable=False,
    )

    # lazy='dynamic' returns a query object instead of collection.
    # Reference: http://docs.sqlalchemy.org/en/latest/orm/collections.html
    roots = relationship(argument="Node", lazy="dynamic", back_populates="namespace")

    users = association_proxy(
        "namespace_users", "user", creator=lambda u: NamespaceUser(user=u)
    )

    user_surrogates = relationship(
        argument="UserSurrogate", lazy="dynamic", back_populates="namespace"
    )

    namespace_owner_requests = relationship(
        argument="NamespaceRequest", lazy="dynamic", back_populates="namespace"
    )

    oauth_records = relationship(
        argument="Oauth", lazy="dynamic", back_populates="namespace"
    )

    watchers = relationship(
        argument="Watcher", lazy="dynamic", back_populates="namespace"
    )

    @property
    def powered_up(self):
        if self.subscription_type in ["production", "trial"]:
            return True
        return False

    def reload_memoized_attr_protection(self):
        """If True, protect attributes from access."""
        if self.powered_up:
            # do not protect attributes, the namespace is powered up.
            self.memoized_attr_protection = False
        else:
            # protect attributes, the namespace is not powered up.
            self.memoized_attr_protection = True

    def __getattribute__(self, attr):
        """Protect certain attributes."""
        # return early if attribute isn't protected.
        if attr not in PROTECTED_ATTRIBUTES:
            return object.__getattribute__(self, attr)

        if not hasattr(self, "memoized_attr_protection"):
            self.reload_memoized_attr_protection()

        if self.memoized_attr_protection:
            # attributes currently under protection.
            # load and return the attribute's default value.
            return PROTECTED_ATTRIBUTES[attr]

        # attributes currently not under protection.
        # load and return the attribute's real value.
        return object.__getattribute__(self, attr)

    def __init__(self, name=None):
        self.id = uuid.uuid1()
        self.name = unicode(name)

    def add_oauth_record(self, user, service, token, data=None):
        self.oauth_records.append(
            Oauth(user=user, namespace=self, service=service, token=token, data=data)
        )

    @property
    def slack_oauth_records(self):
        return [oauth for oauth in self.oauth_records if oauth.service == "slack"]

    def get_namespace_user_for_user(self, user):
        """Given a User object, return the NamespaceUser object."""
        for nsu in self.namespace_users:
            if nsu.user == user:
                return nsu

    def set_role_for_user(self, user, role=None):
        """Add User object to NamespaceUser. Default role is moderator."""
        nsu = self.get_namespace_user_for_user(user)
        if nsu is not None:
            nsu.role = role
        else:
            nsu = NamespaceUser(role=role, namespace=self, user=user)
            self.namespace_users.append(nsu)

        user.watch_namespace(self)

    def disable_user(self, user):
        nsu = self.get_namespace_user_for_user(user)
        if nsu is not None:
            user.unwatch_namespace(self)
            nsu.disabled = True

    def enable_user(self, user):
        nsu = self.get_namespace_user_for_user(user)
        if nsu is not None:
            user.watch_namespace(self)
            nsu.disabled = None

    def remove_user(self, user):
        nsu = self.get_namespace_user_for_user(user)
        if nsu is not None:
            self.disable_user(user)
            self.dbsession.delete(nsu)

    @property
    def enabled_namespace_users(self):
        """Return not disabled NamespaceUser objects for this Namespace."""
        return [nsu for nsu in self.namespace_users if not nsu.disabled]

    @property
    def enabled_users(self):
        """Return not disabled User objects for this Namespace."""
        return [nsu.user for nsu in self.enabled_namespace_users]

    @property
    def roles(self):
        """Return dict with role name as key and list of Users as value."""
        roles = defaultdict(list)
        for ns_user in self.enabled_namespace_users:
            roles[ns_user.role].append(ns_user.user)
        return roles

    @property
    def owners(self):
        """Return a list of owner role User objects."""
        return self.roles.get("owner", [])

    @property
    def moderators(self):
        """Return a list of moderator role User objects."""
        return self.enabled_users

    @property
    def visible_roots(self):
        return self.roots.filter(
            Node.verified == True, Node.disabled == False
        ).order_by(Node.changed.desc())

    @property
    def visible_root_stats(self):
        stats = {}
        for root in self.visible_roots:
            if root.title:
                stats[root.path] = root.stats
            elif root.uri:
                stats[root.uri.data] = root.stats
        return stats

    @property
    def nodes(self):
        return (
            get_nodes_who_share_roots_query(self.dbsession, self.roots)
            .filter(Node.parent_id.isnot(None))
            .order_by(Node.changed.desc())
        )

    def page_nodes(self, limit=100, offset=0):
        if self.roots.count() == 0:
            return []

        query = get_nodes_who_share_roots_query(self.dbsession, self.roots).filter(
            Node.disabled == False, Node.user_id != None
        )

        if self.hide_unless_approved:
            query = query.filter(Node.approved == True)

        if self.hide_unverified:
            query = query.filter(Node.verified == True)

        return query.order_by(Node.changed.desc()).limit(limit).offset(offset)

    @property
    def verified_nodes(self):
        return self.nodes.filter(Node.verified == True)

    @property
    def disabled_nodes(self):
        return self.nodes.filter(Node.disabled == True)

    @property
    def approved_nodes(self):
        return self.nodes.filter(Node.approved == True, Node.disabled == False)

    @property
    def unapproved_nodes(self):
        return self.nodes.filter(
            or_(Node.approved == False, Node.approved.is_(None)), Node.disabled == False
        )

    @property
    def visible_nodes(self):
        return get_nodes_who_share_roots_query(self.dbsession, self.visible_roots)

    @property
    def owner_request_pending(self):
        """Is this Namespace due for a owner request verification?"""
        if self.owner_request_timestamp:
            delta = now_timestamp() - self.owner_request_timestamp
            # 5 minutes.
            if delta <= 300:
                return False
        return True

    @property
    def node_order(self):
        if self.reverse_order:
            return "newest-first"
        return "oldest-first"

    def update_owner_request_timestamp(self):
        self.owner_request_timestamp = now_timestamp()

    def clear_owner_request_timestamp(self):
        self.owner_request_timestamp = None

    def set_stylesheet(self, body):
        self.stylesheet = body
        self.stylesheet_timestamp = now_timestamp()

    def set_stylesheet_embed(self, body):
        self.stylesheet_embed = body
        self.stylesheet_embed_timestamp = now_timestamp()

    def is_owner(self, user):
        """Return True if given user owns this namespace, else False."""
        if user and user.authenticated and user in self.owners:
            return True
        return False

    def is_moderator(self, user):
        """Return True if given user moderates this namespace, else False."""
        if user and user.authenticated and user in self.moderators:
            return True
        return False

    def can_alter_node(self, node, user):
        return self.is_moderator(user) or node.is_owner(user)

    def can_see_node(self, node, user):
        visible = True
        if node.disabled:
            visible = False
        if self.hide_unless_approved and not node.approved:
            visible = False
        if self.hide_unverified and not node.verified:
            visible = False

        if user:
            if self.is_moderator(user) or node.is_owner(user):
                # owners can see node.
                visible = True

        return visible

    @property
    def dict_dump(self):
        dump = {}
        for root in self.roots:
            if root.has_uri:
                root_identifier = root.uri.data
            else:
                root_identifier = root.slug
            root_dump = {
                "name" : root_identifier,
                "title" : root.title,
                # TODO: fix when root.created becomes root.created_timestamp.
                "timestamp" : root.created,
            }
            comments = [] 
            for node in root.children:
                comment = {
                    "date" : node.created_date,
                    # TODO: when root.created becomes root.created_timestamp.
                    "timestamp" : node.created,
                    "content" : node.data,
                    "author" : node.user.name,
                    "author_ip" : node.ip_address,
                }
                if self.subscription_type == "production":
                    comment["email"] = node.user.email
                comments.append(comment)
            root_dump["comments"] = comments
            dump[root_identifier] = root_dump
        return dump


def get_topsecret_namespaces(dbsession):
    return dbsession.query(Namespace).all()


def get_namespace_by_name(dbsession, name):
    if name:
        return (
            dbsession.query(Namespace)
            .filter(Namespace.name == unicode(name))
            .one_or_none()
        )


def get_or_create_namespace(dbsession, name):
    namespace = get_namespace_by_name(dbsession, name)
    if not namespace:
        namespace = Namespace(name)
        # commit namespace to db right away to prevent detached instances.
        # causes some trash namespaces to accumulate but better then errors.
        dbsession.add(namespace)
        dbsession.flush()
    return namespace



"""namespace ignore_query_string

Revision ID: 051f8b810f26
Revises: 3ee195d142fb
Create Date: 2019-02-20 10:27:29.804965

"""

# revision identifiers, used by Alembic.
revision = '051f8b810f26'
down_revision = '3ee195d142fb'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    from sqlalchemy.sql import expression
    op.add_column('rb_namespace', sa.Column('ignore_query_string', sa.Boolean(), nullable=True, server_default=expression.false()))


def downgrade():
    pass



from sqlalchemy.ext.declarative import declarative_base, declared_attr

from sqlalchemy.orm.session import object_session

from sqlalchemy.schema import MetaData
from sqlalchemy import ForeignKey

from time import time

import uuid
from sqlalchemy_utils import UUIDType as TempUUIDType

UUIDType = TempUUIDType(binary=False)

# Recommended naming convention used by Alembic, as various different database
# providers will autogenerate vastly different names making migrations more
# difficult. See: http://alembic.zzzcomputing.com/en/latest/naming.html
NAMING_CONVENTION = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

CLASS_TO_TABLE = {
    "Uri": "rb_uri",
    "User": "rb_user",
    "UserSurrogate": "rb_user_surrogate",
    "Vote": "rb_vote",
    "Node": "rb_node",
    "NodeCache": "rb_node_cache",
    "Namespace": "rb_namespace",
    "NamespaceUser": "rb_namespace_user",
    "NamespaceRequest": "rb_namespace_request",
    "Oauth": "rb_oauth",
    "Watcher": "rb_watcher",
    "NodeEvent": "rb_node_event",
    "NodeEventNotification": "rb_node_event_notification",
    "PayWhatYouCan": "rb_pay_what_you_can",
}

# node (threads), namespace (forum)
WATCHER_TYPES = {"reply", "node", "namespace"}

NOTIFICATION_FREQUENCIES = {"never", "immediately", "daily", "weekly"}

NOTIFICATION_METHODS = {"email"}

NODE_EVENT_ACTIONS = {
    "enabled",
    "disabled",
    "deleted",
    "verified",
    "created",  # this is a new thread.
    "commented",  # this is a reply to an existing thread.
    "approved",
    "edited",
    # These event type actions don't exist yet, but could fuel some ideas for the future.
    # "flagged",
    # "voted",
}

SUBSCRIPTION_TYPES = {"development", "production", "trial"}

# currently not used anywhere.
USER_EVENT_ACTIONS = {"granted", "promoted", "demoted", "blocked"}

metadata = MetaData(naming_convention=NAMING_CONVENTION)
# Base = declarative_base(metadata=metadata)
Base = declarative_base()


def short_id_to_bytes(short_id):
    """Accept a short_id (sanitized url safe base64 string) and return a byte string.

    >>> short_id_to_bytes('dbHeSEFLEeeuz5xONpxxWA')
    'u\xb1\xdeHAK\x11\xe7\xae\xcf\x9cN6\x9cqX'

    """
    return (short_id + "===").replace("_", "/").replace("-", "+").decode("base64")


def id_to_uuid(the_id):
    """
    Accept an id string, return a UUID object.

    >>> id_to_uuid('dbHeSEFLEeeuz5xONpxxWA')
    UUID('75b1de48-414b-11e7-aecf-9c4e369c7158')

    >>> id_to_uuid('75b1de48414b11e7aecf9c4e369c7158')
    UUID('75b1de48-414b-11e7-aecf-9c4e369c7158')

    >>> id_to_uuid('75b1de48-414b-11e7-aecf-9c4e369c7158')
    UUID('75b1de48-414b-11e7-aecf-9c4e369c7158')
    """
    if isinstance(the_id, uuid.UUID):
        # if the_id is already a UUID object, return it.
        return the_id

    try:
        # assume the_id is already a hex uuid string.
        return uuid.UUID(hex=the_id)

    except ValueError:
        try:
            # assume the_id is a base64 "short_id" string.
            return uuid.UUID(bytes=short_id_to_bytes(the_id))
        except ValueError:
            pass

    return None


def get_object_by_id(dbsession, object_id, cls):
    """Try to get object from database by id or return None"""
    object_uuid = id_to_uuid(object_id)
    if object_uuid:
        return dbsession.query(cls).filter(cls.id == object_uuid).one_or_none()


def now_timestamp():
    return int(time() * 1000)


def foreign_key(class_name, column_name):
    return ForeignKey("{}.{}".format(CLASS_TO_TABLE[class_name], column_name))


class RBase(object):
    """Mixin for RemarkBox models."""

    def __eq__(self, other):
        """Determine if equal by id."""
        if other:
            return self.id == other.id
        return False

    def __ne__(self, other):
        """Determine if not equal by user id."""
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    @declared_attr
    def __tablename__(cls):
        return CLASS_TO_TABLE[cls.__name__]

    @property
    def dbsession(self):
        return object_session(self)

    @property
    def id_without_dashes(self):
        return self.id.__str__().replace("-","")



"""create new node_cache table

Revision ID: 0208c10ea67f
Revises: 503a2cef2378
Create Date: 2018-03-28 13:31:34.018012

"""

# revision identifiers, used by Alembic.
revision = "0208c10ea67f"
down_revision = "503a2cef2378"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


from sqlalchemy_utils import UUIDType as TempUUIDType

UUIDType = TempUUIDType(binary=False)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "rb_node_cache",
        sa.Column("id", UUIDType, nullable=False),
        sa.Column("node_id", UUIDType, nullable=True),
        sa.Column("valid", sa.Boolean(), nullable=True),
        sa.Column("updated_timestamp", sa.BigInteger(), nullable=False),
        sa.Column("json_stats", sa.UnicodeText(), nullable=True),
        sa.ForeignKeyConstraint(["node_id"], ["rb_node.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_rb_node_cache_id"), "rb_node_cache", ["id"], unique=False)
    op.create_index(
        op.f("ix_rb_node_cache_node_id"), "rb_node_cache", ["node_id"], unique=False
    )
    op.drop_table("rb_root_cache")


def downgrade():
    op.create_table(
        "rb_root_cache",
        sa.Column("id", sa.CHAR(length=32), nullable=False),
        sa.Column("root_id", sa.CHAR(length=32), nullable=True),
        sa.Column("valid", sa.BOOLEAN(), nullable=True),
        sa.Column("updated_timestamp", sa.BIGINT(), nullable=False),
        sa.Column("json_stats", sa.TEXT(), nullable=True),
        sa.CheckConstraint(u"valid IN (0, 1)"),
        sa.ForeignKeyConstraint(["root_id"], [u"rb_node.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.drop_index(op.f("ix_rb_node_cache_node_id"), table_name="rb_node_cache")
    op.drop_index(op.f("ix_rb_node_cache_id"), table_name="rb_node_cache")
    op.drop_table("rb_node_cache")



from pyramid.view import view_config

from pyramid.csrf import check_csrf_token

from pyramid.httpexceptions import HTTPFound

from . import (
    get_referer_or_home,
    get_embed_route_uri,
    get_node_route_uri,
    set_node_to_pending_in_session,
)

from remarkbox.lib.notify import schedule_notifications

try:
    unicode("")
except:
    from six import u as unicode


# check CSRF only if user is authenticated.
@view_config(route_name="embed-reply", require_csrf=False)
@view_config(route_name="embed-reply2", require_csrf=False)
@view_config(route_name="basic-reply", require_csrf=False)
@view_config(route_name="basic-reply2", require_csrf=False)
def reply_node(request):
    """handle posting of reply form from show-node pages."""
    thread_data = request.params.get("thread_data", "")

    # return early if spam attribute is truthy.
    if request.spam:
        return request.spam

    # return early if node (parent) not not exist.
    if request.node is None:
        # redirect to the referer or home. No node found in database.
        return HTTPFound(get_referer_or_home(request))

    # return early if node (parent) is disabled.
    if request.node.disabled is True:
        # redirect to the referer or home.
        request.session.flash(("No remarks for the disabled.", "error"))
        return HTTPFound(get_referer_or_home(request))

    # flash error and return early if user is None.
    if request.user is None:
        request.session.flash(
            ("Press the back button to fix your email address", "error")
        )
        return HTTPFound(get_referer_or_home(request))

    # return early if thread (root node) is locked.
    if request.node.root.locked:
        request.session.flash(
            ("You may not comment, this thread is locked.", "error")
        )
        return HTTPFound(get_referer_or_home(request))


    # check CSRF only if user is authenticated.
    if request.method == "POST" and request.csrf_token:
        check_csrf_token(request)

    # flash error and return early if data is empty.
    if thread_data == "":
        request.session.flash(("Your message was empty", "error"))
        return HTTPFound(get_referer_or_home(request))

    # STEP 1: get a parent node.
    parent = request.node

    # STEP 2: attach a brand new child node to parent node.
    child = parent.new_child()
    child.user = request.user
    child.ip_address = unicode(request.client_addr)
    child.verified = request.user.authenticated
    child.set_data(thread_data, namespace=request.namespace)
    child_event = child.new_event(request.user, "commented")

    if request.namespace.hide_unless_approved:
        # by default comments are approved, unless Namespace hide_unless_approved
        # is enabled, moderators nodes are always auto approved.
        child.approved = request.namespace.is_moderator(request.user)

    # STEP 3: update root's changed timestamp.
    # TODO: maybe we should find a better way to "bump" a thread.
    parent.root.changed = child.changed
    # invalidate root cache to force recomputation.
    parent._invalidate_cache()

    # STEP 4: commit to database.
    request.dbsession.add(request.user)
    request.dbsession.add(child)
    request.dbsession.add(child_event)
    request.dbsession.add(parent)
    request.dbsession.add(parent.root)
    request.dbsession.flush()

    schedule_notifications(request, child_event)

    msg = ("Your post was successful!", "success")
    request.session.flash(msg)

    ### TODO: everything below this is pretty much crap code...
    #         and likely deserves a flowchart...

    # set return_to URI.
    if request.mode == "embed":
        return_to = get_embed_route_uri(request, child.root.uri.data, child.id)
    else:
        return_to = get_node_route_uri(request, child.root, child.id)

    if child.verified == True:
        # Redirect to new node if user and new node is verified.
        return HTTPFound(return_to)

    set_node_to_pending_in_session(request, child)

    if request.mode == "embed":
        # set return_to to the parent iframe URI.
        return_to = "{}#{}".format(child.root.uri.data, child.id)

    # Redirect to join-or-log-in, posting email and submit.
    # Per the docs: " Extra replacement names are ignored."
    # In this case "namespace" is extra when mode is basic.
    # https://docs.pylonsproject.org/projects/pyramid/en/latest/api/request.html#pyramid.request.Request.route_url
    uri = request.route_url(
        route_name=request.mode + "-join-or-log-in",
        namespace=request.namespace.name,
        _query={"email": request.user.email, "return-to": return_to, "submit": True},
    )
    return HTTPFound(uri)



from pyramid.view import view_config

from remarkbox.views import user_required

from remarkbox.models.notification import NodeEventNotification


@view_config(route_name="basic-user-notifications", renderer="list-notifications.j2")
@view_config(route_name="embed-user-notifications", renderer="list-notifications.j2")
@user_required()
def user_notifications(request):
    if "all" in request.params: 
        state = "all"
        notifications = request.user.node_notifications
    elif "sent" in request.params: 
        state = "sent"
        notifications = request.user.sent_node_notifications
    else:
        state = "unsent"
        notifications = request.user.unsent_node_notifications

    paginated_notifications = (
        notifications
            .order_by(
                NodeEventNotification.created_timestamp.desc(),
                NodeEventNotification.user_id,
            )
            .offset(request.page_offset)
            .limit(request.page_size)
    )

    notification_counts = {
        "all": request.user.node_notifications.count(),
        "sent": request.user.sent_node_notifications.count(),
        "unsent": request.user.unsent_node_notifications.count(),
    }

    return {
        "notifications": paginated_notifications,
        "the_title": "Notifications",
        "topsecret": False,
        "state": state,
        "counts": notification_counts,
    }



"""namespace reverse node order

Revision ID: f7b42b2024fb
Revises: 40ded1163f95
Create Date: 2019-03-19 11:17:31.427157

"""

# revision identifiers, used by Alembic.
revision = 'f7b42b2024fb'
down_revision = '40ded1163f95'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('rb_namespace', sa.Column('reverse_order', sa.Boolean(), nullable=True))

def downgrade():
    pass




defaults = {
    "size": 100,
    "font_size": 60,
    "font_family": "Arial",
    "bg": "black",
    "color": "white",
    "text": "KB",
}


def make_svg(**options):
    """Builds a simple SVG text square with a centered text.

    options: size, font_size, font_family, bg, color, text

    :return (str)
    """
    if options is not None and not isinstance(options, dict):
        raise ValueError("options should be None or a type of dict.")

    if "size" in options and "font_size" not in options:
        options["font_size"] = int(options["size"] * 0.6)

    dc = defaults.copy()
    dc.update(options)
    options = dc

    svg = u"""
    <svg xmlns="http://www.w3.org/2000/svg"
         width="{size}px" height="{size}px">
      <g>
        <rect x="0" y="0" fill="{bg}" width="{size}px" height="{size}px">
        </rect>
        <text y="50%" x="50%" fill="{color}"
              text-anchor="middle" dominant-baseline="central"
              style="font-family: {font_family}; font-size: {font_size}px">
          {text}
        </text>
      </g>
    </svg>
    """.format(
        **options
    )

    return svg



from pyramid.paster import bootstrap, setup_logging

from ..models import get_namespace_by_name

from . import base_parser


def get_arg_parser():
    parser = base_parser("Modify all uris which fall under namespace.")
    parser.add_argument("namespace")
    parser.add_argument("--add-trailing-slash", action="store_true", default=False)
    parser.add_argument("--https", action="store_true", default=False)
    parser.add_argument("--http", action="store_true", default=False)
    parser.add_argument("--move-namespace-domain")
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        namespace = get_namespace_by_name(request.dbsession, args.namespace)
        if namespace is not None:

            for root in namespace.roots:
                if root.uri is None:
                    continue

                # parsed miniuri object.
                puri = root.uri.parsed

                if args.add_trailing_slash and not root.uri.data.endswith("/"):
                    print("adding trailing slash to: {}".format(root.uri.data))
                    puri.path = puri.path + "/"

                if args.move_namespace_domain and "." in args.move_namespace_domain:
                    namespace.name = args.move_namespace_domain
                    puri.hostname = args.move_namespace_domain

                if args.http:
                    puri.scheme = "http"

                if args.https:
                    puri.scheme = "https"

                root.uri.data = puri.uri
                request.dbsession.add(root.uri)

            request.dbsession.flush()
        else:
            print("no such namespace ({})".format(args.namespace))



from markdown import markdown

from functools import partial

from collections import defaultdict

from bleach.sanitizer import Cleaner
from bleach.linkifier import LinkifyFilter
from bleach.callbacks import nofollow, target_blank

from bleach_allowlist import markdown_tags, markdown_attrs, all_styles

from bs4 import BeautifulSoup

import miniuri

import logging

log = logging.getLogger(__name__)


def default_tag_acl():
    return defaultdict(list)


def default_cleaner(tag_acl=None):
    """
    Returns a default Cleaner object.

    We use BeautifulSoup to conditionally whitelist or blacklist
    tags based on tag attr_name and attr_value pairs.

    For example, this will whitelist Mathjax script tags:

    tag_acl = {
        "script": [
            ("type", "math/tex; mode=display", "allow"),
        ],
    }

    While this example will blacklist Mathjax script tags:

    tag_acl = {
        "script": [
            ("type", "math/tex; mode=display", "deny"),
        ],
    }

    """
    if tag_acl is None:
        tag_acl = {}

    maybe_safe_tags = ["pre", "table", "tr", "td"]

    tags = maybe_safe_tags + list(tag_acl.keys()) + markdown_tags
    attrs = markdown_attrs

    attrs["img"].append("width")
    attrs["span"] = ["class"]

    # Allow both whitelist and blacklist tag_name/attr/attr_value
    # to get past bleach.
    # We will conditionally filter tags using BeautifulSoup.
    for tag_name in tag_acl.keys():

        # tag_name, for example "script".

        for attr_name, attr_value, allow_or_deny in tag_acl[tag_name]:

            if allow_or_deny == "allow":

                # attr_name, for example "type".
                # attr_value, for example "math/tex; mode=display".

                if tag_name not in attrs:
                    # if tag_name not in attrs, create it as an empty list.
                    attrs[tag_name] = []

                # append the attr_name to the map of approved
                # attributes for the given tag_name.
                attrs[tag_name].append(attr_name)

    cleaner = Cleaner(tags=tags, attributes=attrs)

    # doesn't do anything, but i used to be able to pass it via constructor.
    # https://github.com/yourcelf/bleach-allowlist/blob/main/bleach_allowlist/bleach_allowlist.py
    #cleaner.all_styles=all_styles

    # disable link_protection by default.
    cleaner.link_protection = False
    # an None signifies a relative URI, which should always be whitelisted.
    cleaner.whitelist_domains = [None]
    # absolute domain, used for turning relative paths into absolute.
    cleaner.absolute_domain = ""
    # add conditional_whitelist_tags to cleaner object.
    cleaner.tag_acl = tag_acl

    return cleaner


def markdown_to_raw_html(data, extra_extensions=None):
    """Accepts a markdown string, returns raw unsanitized HTML"""
    extensions = [
        "markdown.extensions.codehilite",
        "markdown.extensions.fenced_code",
    ]
    if extra_extensions is not None:
        extensions.extend(extra_extensions)

    return markdown(data, extensions=extensions)


def conditional_tag_filter(soup, cleaner):
    """
    Use BeautifulSoup `Soup` object to filter out non-whitelisted tags.
    """

    # bleach does not have a way to conditionally accept tags whitelists.
    # so we built our own way.

    log.info(cleaner.tag_acl)

    # for each tag_name in the tag_acl ...
    for tag_name in cleaner.tag_acl.keys():

        # find all the tags in the DOM that match the tag_name, for example "script".
        for tag in soup.find_all(tag_name):

            # tag is a BeautifulSoup tag object.

            # by default, assume we will extract tag (remove it from the DOM).
            extract_tag = True

            for attr_name, attr_value, allow_or_deny in cleaner.tag_acl[tag_name]:

                # tag attr_name:
                #
                #   * "type"
                #   * "id"
                #   * "class"
                #   * ect
                #
                # tag attr_value:
                #
                #   * "math/tex; mode=display"
                #   * "<uuid>"
                #   * ect
                #
                # allow_or_deny:
                #
                #  * "allow"
                #  * "deny"

                log.info("{} {} {}".format(attr_name, attr_value, allow_or_deny))

                if allow_or_deny == "allow":

                    if tag.attrs.get(attr_name) == attr_value:
                        # this tag attr_name/attr_value is whitelisted.
                        # do not extract tag (remove it from the DOM)
                        extract_tag = False
                        break

            if extract_tag:
                log.info(
                    "Extracting {} tag with attr_name {} and attr_value {}".format(
                        tag_name, attr_name, attr_value
                    )
                )
                # remove this tag.
                # this tag attr_name:attr_value is _not_ whitelisted.
                tag.extract()

    return soup


def protect_links(soup, cleaner):

    for a_tag in soup.find_all("a"):
        uri = miniuri.Uri(a_tag.attrs.get("href", ""))

        if uri.hostname in cleaner.whitelist_domains:
            # domain in whitelist or relative URI so remove rel="nofollow".
            a_tag.attrs.pop("rel", None)
            if uri.hostname is None and cleaner.absolute_domain:
                # make relative path absolute!
                # TODO: if we hold scheme in Namespace object we can use it
                # when building absolute uris. for now assume https.
                uri.scheme = "https"
                uri.hostname = cleaner.absolute_domain
                a_tag.attrs["href"] = str(uri)

        elif cleaner.link_protection:
            # domain not in whitelist, replace a_tag with "[link removed]".
            a_tag.replace_with("[link removed]")

    return soup


def clean_raw_html(raw_html, cleaner=None):
    """
    Accepts raw HTML and a cleaner object
    Returns sanitized HTML as bytes.
    """
    if cleaner is None:
        cleaner = default_cleaner()

    if cleaner.link_protection == False:
        cleaner.filters.append(
            partial(
                LinkifyFilter,
                callbacks=[nofollow, target_blank],
                skip_tags=["pre", "code"],
            )
        )

    cleaned_html = cleaner.clean(raw_html)

    soup = BeautifulSoup(cleaned_html, "html5lib")

    # conditionally accept whitelisted tags, filter out the rest.
    soup = conditional_tag_filter(soup, cleaner)

    # protect links from abuse.
    soup = protect_links(soup, cleaner)

    return soup.decode(eventual_encoding="utf-8")



"""hide_powered_by column

Revision ID: aa2dfb4decda
Revises: 4aeda8e797ab
Create Date: 2018-03-15 15:44:31.591391

"""

# revision identifiers, used by Alembic.
revision = "aa2dfb4decda"
down_revision = "4aeda8e797ab"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_namespace", sa.Column("hide_powered_by", sa.Boolean(), nullable=True)
    )


def downgrade():
    op.drop_column("rb_namespace", "hide_powered_by")



"""add root_id field

Revision ID: c9793937e756
Revises: 051f8b810f26
Create Date: 2017-01-14 20:59:39.701946

"""

# revision identifiers, used by Alembic.
revision = "c9793937e756"
down_revision = "051f8b810f26"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import NamespaceRequest


def upgrade():

    bind = op.get_bind()
    DBSession = sa.orm.Session(bind=bind)
    namespace_requests = DBSession.query(NamespaceRequest).all()

    for namespace_request in namespace_requests:
        if namespace_request.verified:
            namespace = namespace_request.namespace
            namespace.set_role_for_user(namespace_request.user, "owner")
            DBSession.add(namespace)

    DBSession.flush()
    DBSession.commit()


def downgrade():
    pass



import re

from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound

from remarkbox.models import get_node_by_uri, get_node_by_id

from remarkbox.models.user import get_or_create_user_by_email

from remarkbox.lib.mail import send_verification_digits_to_email

from . import get_referer_or_home, get_embed_route_uri, verify_pending_nodes_in_session

from urllib.parse import urlencode


@view_config(route_name="log-out")
@view_config(route_name="embed-log-out")
def log_out(request):
    """log out the user, redirect to back to referer."""
    uri = get_referer_or_home(request)
    request.session["authenticated_user_id"] = None
    return HTTPFound(uri)


#@view_config(route_name="join-or-log-in", renderer="join-or-log-in.j2")
@view_config(route_name="basic-join-or-log-in", renderer="join-or-log-in.j2", require_csrf=False)
@view_config(route_name="embed-join-or-log-in", renderer="join-or-log-in.j2", require_csrf=False)
def join_or_log_in(request):
    """
    This view handles user registration, verification, and log in.
    It uses "password-less" authentication by sending 6 digit
    OTP (one-time-password) tokens to email addresses to verify both the email
    address & to authenticate the device displaying the challenge input field.
    """
    _email_regex = re.compile("^[^@]+@[^@]+\.[^.@]+$")

    # get the raw OTP (one-time-password) from posted parameters.
    raw_otp = request.params.get("raw-otp", "")

    # get the email from posted parameters.
    email = request.params.get("email", "")

    if email and _email_regex.match(email) is None:
        # posted email does not pass regex, set it to None.
        email = ""
        request.session.flash(("That email address is invalid.", "error"))

    if request.spam:
        return request.spam

    if request.user and request.user.authenticated:
        #return HTTPFound(get_referer_or_home(request))
        return HTTPFound("/")

    elif email:
        # get or create a User object from the posted email.
        user = get_or_create_user_by_email(request.dbsession, email)

        if user.throttle_password():
            msg = (
                "We already sent a link to {}. Check email to log in.".format(user.email),
                "info",
            )

        else:
            # generate a new one-time-password and save to database
            raw_otp = user.new_password()
            request.dbsession.add(user)
            request.dbsession.flush()

            # email user the one-time-password and flash message.
            send_verification_digits_to_email(request, user.email, raw_otp)

            msg = (
                "We just sent a link to {}. Check email to log in.".format(user.email),
                "info",
            )

        request.session.flash(msg)

        email_encoded = urlencode({"email": email})

        return HTTPFound("{}/verification-challenge?{}".format(request.link_prefix, email_encoded))

    return {
        "title": "join or log in",
    }


@view_config(route_name="basic-verification-challenge", renderer="verification-challenge.j2", require_csrf=False)
@view_config(route_name="embed-verification-challenge", renderer="verification-challenge.j2", require_csrf=False)
def verification_challenge(request):

    # get the raw OTP (one-time-password) from posted parameters.
    raw_otp = request.params.get("raw-otp", "")

    # get the email from posted parameters.
    email = request.params.get("email", "")

    user = None
    if email:
        # get or create a User object from the posted email.
        user = get_or_create_user_by_email(request.dbsession, email)

    if raw_otp:
        if user.check_password(raw_otp):
            # success: the user was verified.
            user.verified = True
            msg = ("Welcome {}".format(user.name), "success")
            request.session["authenticated_user_id"] = str(user.id)
            request.session.flash(msg)

            # attempt to verify all nodes_pending_verify in user's session.
            verify_pending_nodes_in_session(request, user)

            # Idempotent operation. Make certain a user has at least one reply_watcher.
            user.create_default_reply_watcher()

            request.dbsession.add(user)
            request.dbsession.flush()

            return HTTPFound("{}/u/settings".format(request.link_prefix))

        else:
            msg = ("Invalid Verification Code", "error")
            request.session.flash(msg)

    return {
        "title": "Please Enter Verification Code",
        "email": email,
        "raw_otp": raw_otp,
    }




from .meta import WATCHER_TYPES, NOTIFICATION_FREQUENCIES, NOTIFICATION_METHODS
from .meta import Base, RBase, UUIDType, foreign_key, get_object_by_id, now_timestamp

from sqlalchemy import Column, BigInteger, Unicode, UnicodeText, Enum

from sqlalchemy.orm import relationship

import uuid

from .notification import NodeEventNotification


class Watcher(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    # the User watching the Node.
    user_id = Column(UUIDType, foreign_key("User", "id"), index=True)
    # the Node under watch.
    node_id = Column(UUIDType, foreign_key("Node", "id"), index=True, nullable=True)
    # the Namespace under watch.
    namespace_id = Column(
        UUIDType, foreign_key("Namespace", "id"), index=True, nullable=True
    )
    # the type of watcher object.
    type = Column(Enum(*WATCHER_TYPES, name="type"))
    # how will this watcher object notify?
    notification_method = Column(
        Enum(*NOTIFICATION_METHODS, name="notification_method"), default="email"
    )
    # how frequently should we notify of changes?
    notification_frequency = Column(
        Enum(*NOTIFICATION_FREQUENCIES, name="frequency"), default="daily"
    )
    # the point in time when the User started authorized consent to watch.
    created_timestamp = Column(BigInteger, nullable=False)
    # the last time this record was modified.
    updated_timestamp = Column(BigInteger, nullable=False)

    def __init__(
        self,
        user,
        type,
        notification_method="email",
        notification_frequency="daily",
        node=None,
        namespace=None,
    ):
        self.id = uuid.uuid1()
        self.user = user
        self.type = type
        self.notification_method = notification_method
        self.notification_frequency = notification_frequency
        self.created_timestamp = now_timestamp()
        self.updated_timestamp = now_timestamp()

        self.node = node
        self.namespace = namespace

    # 1-to-1 relationships.
    node = relationship(argument="Node", uselist=False, lazy="joined")
    namespace = relationship(argument="Namespace", uselist=False, lazy="joined")
    user = relationship(argument="User", uselist=False, lazy="joined")
    notifications = relationship(
        argument="NodeEventNotification",
        lazy="dynamic",
        cascade="save-update, merge, delete",
        back_populates="watcher"
    )

    def unsent_notifications(self):
        self.notifications.filter(NodeEventNotification.sent == False)

    def new_notification(self, node_event):
        notification = NodeEventNotification(watcher=self, node_event=node_event)
        self.notifications.append(notification)
        return notification


def get_watcher_by_id(dbsession, watcher_id):
    """Try to get Watcher object by id or return None."""
    return get_object_by_id(dbsession, watcher_id, Watcher)



"""stylesheet_timestamp

Revision ID: 59592cf325fb
Revises: 256468780798
Create Date: 2017-07-06 00:30:42.512137

"""

# revision identifiers, used by Alembic.
revision = "59592cf325fb"
down_revision = "256468780798"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "Namespace", sa.Column("stylesheet_timestamp", sa.BigInteger(), nullable=True)
    )


def downgrade():
    op.drop_column("Namespace", "stylesheet_timestamp")



from sqlalchemy import BigInteger, Boolean, Integer, Column, Unicode, Enum, func, or_

from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.associationproxy import association_proxy

import uuid

import bcrypt

from remarkbox.lib import timestamp_to_ago_string, gravatar_client

from remarkbox.lib.svgtar import inline_svgtar

from .meta import NOTIFICATION_FREQUENCIES
from .meta import Base, RBase, UUIDType, now_timestamp, foreign_key, get_object_by_id

from .node import Node

from .namespace_request import NamespaceRequest

from .watcher import Watcher

from .notification import NodeEventNotification

import logging

log = logging.getLogger(__name__)

try:
    unicode("")
except:
    from six import u as unicode


def generate_password(size=32):
    """Return a system generated password"""
    from random import choice
    letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    digits = "0123456789"
    pool = letters + digits
    return "".join([choice(pool) for i in range(size)])


class UserSurrogate(RBase, Base):
    """
    This class represents a user surrogate account, a stand in.

    A user surrogate may not log in; has no email. It is not a "real"
    User, but a surrogate, a ghost of a user imported from another system.
    """

    id = Column(UUIDType, primary_key=True, index=True)
    # the namespace which imported this user surrogate.
    namespace_id = Column(UUIDType, foreign_key("Namespace", "id"), index=True)
    # the display name.
    name = Column(Unicode(64), nullable=False)
    created_timestamp = Column(BigInteger, nullable=False)

    namespace = relationship(argument="Namespace", uselist=False, lazy="joined")

    # lazy='dynamic' returns a query object instead of collection.
    # Reference: http://docs.sqlalchemy.org/en/latest/orm/collections.html
    nodes = relationship(
        argument="Node",
        lazy="dynamic",
        order_by=Node.created.desc(),
        back_populates="user_surrogate",
    )

    def __init__(self, name, namespace):
        self.name = name
        self.namespace_id = namespace.id
        self.created_timestamp = now_timestamp()
        self.id = uuid.uuid1()

    def svgtar(self, **kwargs):
        if "text" not in kwargs:
            kwargs["text"] = self.name[:1]
        return inline_svgtar(h=str(self.id), **kwargs)

    def avatar_uri(self, **kwargs):
        return self.svgtar(**kwargs)


class User(RBase, Base):
    """This class represents a user account"""

    # was this User object authenticated?
    authenticated = False

    id = Column(UUIDType, primary_key=True, index=True)
    name = Column(Unicode(64), unique=True, nullable=False)
    email = Column(Unicode(64), unique=True, nullable=False)
    # the email_id is a random string added to otp links to prevent an
    # attacker from looping/triggering otp emails for a list of addresses.
    # we only share this string with the owner of an email box. not-in-use...
    email_id = Column(Unicode(8))
    password = Column(Unicode(64))
    password_attempts = Column(Integer, default=0)
    password_timestamp = Column(BigInteger)
    # TODO: someday this should be renamed to created_timestamp
    created = Column(BigInteger, nullable=False)
    gravatar = Column(Boolean, default=False)
    verified = Column(Boolean, default=False)
    disabled = Column(Boolean, default=False)
    # automatically watch any threads I create.
    auto_watch_threads_i_create = Column(Boolean, default=True, nullable=False)
    # automatically watch any threads I participate in.
    auto_watch_threads_i_participate = Column(Boolean, default=False, nullable=False)
    # By default, when I watch a Thread, I want to be notified:
    default_node_watcher_frequency = Column(
        Enum(*NOTIFICATION_FREQUENCIES, name="frequency"),
        default="daily",
        nullable=False,
    )
    # example: cus_12345678AbCdEF but may be null.
    stripe_id = Column(Unicode(18), unique=True, nullable=True)

    votes = relationship(argument="Vote", backref="user", order_by="desc(Vote.created)")

    # lazy='dynamic' returns a query object instead of collection.
    # Reference: http://docs.sqlalchemy.org/en/latest/orm/collections.html
    nodes = relationship(
        argument="Node",
        lazy="dynamic",
        order_by=Node.created.desc(),
        back_populates="user",
    )

    namespaces = association_proxy(
        "user_namespaces", "namespace", creator=lambda n: NamespaceUser(namespace=n)
    )

    oauth_records = relationship(
        argument="Oauth", lazy="dynamic", back_populates="user"
    )

    namespace_owner_requests = relationship(
        argument="NamespaceRequest", lazy="dynamic", back_populates="user"
    )

    watchers = relationship(
        argument="Watcher",
        lazy="dynamic",
        back_populates="user",
        order_by=Watcher.created_timestamp,
        cascade="save-update, merge, delete",
    )

    # 1-to-1 relationships.
    pay_what_you_can = relationship(argument="PayWhatYouCan", uselist=False, lazy="joined")

    @property
    def node_watchers(self):
        return self.watchers.filter(Watcher.type == "node")

    @property
    def namespace_watchers(self):
        return self.watchers.filter(Watcher.type == "namespace")

    @property
    def reply_watchers(self):
        return self.watchers.filter(Watcher.type == "reply")

    @property
    def reply_watcher(self):
        """for now we only support one reply_watcher in the UI."""
        return self.watchers.filter(Watcher.type == "reply").one()

    def get_watcher_by_id(self, watcher_id):
        """Return a Watcher object for by ID or None."""
        return self.node_watchers.filter(Watcher.id == watcher_id).one_or_none()

    def get_watcher_by_node_id(self, node_id):
        """Return a Watcher object for the node (thread) or None."""
        return self.node_watchers.filter(Watcher.node_id == node_id).one_or_none()

    def watch_node(self, node):
        """create and return new Watcher object for this node and user in memory"""
        watcher = Watcher(
            user=self,
            type="node",
            node=node,
            notification_frequency=self.default_node_watcher_frequency,
        )
        self.watchers.append(watcher)
        return watcher

    def watch_namespace(self, namespace):
        """create and return new Watcher object for this namespace and user in memory"""
        watcher = self.get_watcher_by_namespace_id(namespace.id)
        if watcher:
            return watcher

        watcher = Watcher(
            user=self,
            type="namespace",
            namespace=namespace,
            notification_frequency="daily",
        )
        self.watchers.append(watcher)
        return watcher

    def get_watcher_by_namespace_id(self, namespace_id):
        """Return a Watcher object for the Namespace or None."""
        return self.namespace_watchers.filter(
            Watcher.namespace_id == namespace_id
        ).one_or_none()

    def unwatch_namespace(self, namespace):
        """Stop watching a Namespace."""
        watcher = self.get_watcher_by_namespace_id(namespace.id)
        if watcher:
            self.dbsession.delete(watcher)
            self.dbsession.flush()

    node_notifications = relationship(
        argument="NodeEventNotification", lazy="dynamic", back_populates="user"
    )

    @property
    def unsent_node_notifications(self):
        return self.node_notifications.filter(NodeEventNotification.sent == False)

    @property
    def sent_node_notifications(self):
        return self.node_notifications.filter(NodeEventNotification.sent == True)

    @property
    def verified_namespace_owner_requests(self):
        return [nr for nr in self.namespace_owner_requests if nr.verified]

    @property
    def unverified_namespace_owner_requests(self):
        return [nr for nr in self.namespace_owner_requests if not nr.verified]

    @property
    def verified_nodes(self):
        return self.nodes.filter(Node.verified == True, Node.disabled == False)

    @property
    def unverified_nodes(self):
        return self.nodes.filter(Node.verified == False, Node.disabled == False)

    @property
    def disabled_nodes(self):
        return self.nodes.filter(Node.verified == True, Node.disabled == True)

    @property
    def unapproved_nodes(self):
        return self.nodes.filter(or_(Node.approved == False, Node.approved.is_(None)))

    def page_nodes(self, limit=100, offset=0):
        if self.nodes.count() == 0:
            return []
        return (
            self.nodes.filter(
                Node.disabled == False, Node.verified == True, Node.user_id != None, Node.approved == True
            )
            .order_by(Node.changed.desc())
            .limit(limit)
            .offset(offset)
        )

    def __init__(self, email):
        self.name = unicode(generate_password(size=8))
        self.created = now_timestamp()
        self.id = uuid.uuid1()
        self.email = unicode(email)

        # TODO: this field was never used & should be sunset.
        self.email_id = unicode(generate_password(size=8))

        self.new_password()
        # don't password throttle new User objects.
        self.password_timestamp = 0

    def _generate_raw_password(self):
        """Return a system generated password"""
        #return generate_password(size)
        from random import choice
        numbers = '0123456789'
        return "".join([choice(numbers) for i in range(0,6)])

    def new_password(self):
        """Generate and return raw password, store password hash into DB."""
        raw_password = self._generate_raw_password()

        # bcrypt works with bytes so we encode to utf-8.
        self.password = bcrypt.hashpw(
            raw_password.encode("utf-8"),
            bcrypt.gensalt()
        ).decode("utf-8")
        self.password_timestamp = now_timestamp()
        self.password_attempts = 0
        return raw_password

    def check_password(self, password):
        """Accept plain-text raw password, create hash, compare with DB."""
        stored_hash = self.password

        # increment password attempts.
        self.password_attempts += 1

        # expire password after 15 minutes.
        # 900000 milliseconds == 15 minutes
        if self.password_timestamp_delta >= 900000:
            return False

        # prevent brute force, allow 10 invalid verification code attempts.
        if self.password_attempts >= 10:
            return False

        # bcrypt works with bytes so we encode to utf-8.
        new_hash = bcrypt.hashpw(
            password.encode("utf-8"),
            stored_hash.encode("utf-8"),
        ).decode("utf-8")

        log.info("new_hash={} stored_hash={}".format(new_hash, stored_hash))

        if new_hash == stored_hash:
            return True
        return False

    def throttle_password(self, needed_delta=90000):
        """Return True when throttled, else False"""
        if self.password_timestamp_delta <= needed_delta:
            return True
        return False

    @property
    def password_timestamp_delta(self):
        return now_timestamp() - self.password_timestamp

    @property
    def human_password_timestamp(self):
        return timestamp_to_ago_string(self.password_timestamp)

    @property
    def human_timestamp(self):
        return timestamp_to_ago_string(self.created)

    def gravatar_uri(self, **kwargs):
        return gravatar_client(self.email, **kwargs)

    def svgtar(self, **kwargs):
        if "text" not in kwargs:
            kwargs["text"] = self.name[:1]
        return inline_svgtar(h=str(self.id), **kwargs)

    def avatar_uri(self, **kwargs):
        if self.gravatar:
            return self.gravatar_uri(**kwargs)
        return self.svgtar(**kwargs)

    def create_default_reply_watcher(self):
        if self.reply_watchers.count() == 0:
            reply_watcher = Watcher(self, "reply")
            self.watchers.append(reply_watcher)
            return reply_watcher
        return self.reply_watchers.first()


def _user_by_name_query(dbsession, name):
    """query User by case insensitive name."""
    return dbsession.query(User).filter(func.lower(User.name) == unicode(name.lower()))


def is_user_name_available(dbsession, name):
    return not _user_by_name_query(dbsession, unicode(name)).count()


def generate_user_name(dbsession, desired_name=""):
    size = 8
    sep = ""
    if desired_name:
        size = 3
        sep = "-"
    name = sep.join([desired_name, generate_password(size)])
    while is_user_name_available(dbsession, name) == False:
        name = sep.join([desired_name, generate_password(size)])
    return name


def is_user_name_valid(name):
    """Test if user name meets our criteria for validity."""
    # The choices here are subject to change.  Right now we allow dashes.
    name = name.replace("-", "")
    return name.isalnum()


def get_user_by_name(dbsession, name):
    """Try to get User object by name or return None"""
    if name:
        return _user_by_name_query(dbsession, name).one_or_none()


def get_user_by_id(dbsession, user_id):
    """Try to get User object by id or return None."""
    return get_object_by_id(dbsession, user_id, User)


def get_user_by_email(dbsession, email):
    """Try to get User object by email or return None"""
    if email:
        return dbsession.query(User).filter(User.email == unicode(email)).one_or_none()


def get_or_create_user_by_email(dbsession, email):
    """Try to get User object by email or return new User"""
    user = get_user_by_email(dbsession, email)
    if user is None:
        # create new User for unverified user.
        user = User(email)
        if not is_user_name_available(dbsession, user.name):
            user.name = generate_user_name(dbsession)
    return user


def get_user_surrogate_by_name(dbsession, name, namespace):
    """Try to get UserSurrogate object by case insensitive name or return None"""
    if name:
        return (
            dbsession.query(UserSurrogate)
            .filter(func.lower(User.name) == unicode(name.lower()))
            .filter(UserSurrogate.namespace_id == namespace.id)
            .one_or_none()
        )


def get_or_create_user_surrogate_by_name(dbsession, name, namespace):
    """
    Try to get UserSurrogate object by case insensitive name
    or create and return a new UserSurrogate
    """
    user_surrogate = get_user_surrogate_by_name(dbsession, name, namespace)
    if user_surrogate is None:
        user_surrogate = UserSurrogate(name=name, namespace=namespace)
    return user_surrogate



"""namespace model

Revision ID: 98a90d2add21
Revises: 04e52337756e
Create Date: 2017-07-03 16:13:09.160494

"""

# revision identifiers, used by Alembic.
revision = "98a90d2add21"
down_revision = "04e52337756e"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import UUIDType


def upgrade():
    # we don't need to create the table.
    # it will be created automatically during release.
    # We do need to create the namespace_id column on the node table.
    op.add_column(u"Node", sa.Column("namespace_id", UUIDType, nullable=True))

    # this fails so I had to comment it out.
    # op.create_foreign_key(None, 'Node', 'Namespace', ['namespace_id'], ['id'])


def downgrade():
    op.drop_constraint(None, "Node", type_="foreignkey")
    op.drop_column(u"Node", "namespace_id")
    op.drop_index(op.f("ix_Namespace_name"), table_name="Namespace")
    op.drop_index(op.f("ix_Namespace_id"), table_name="Namespace")
    op.drop_table("Namespace")



"""notification columns to rb_user table

Revision ID: 6fc1d0cf350c
Revises: a26302bcfa25
Create Date: 2018-05-27 21:04:17.704776

"""

# revision identifiers, used by Alembic.
revision = "6fc1d0cf350c"
down_revision = "a26302bcfa25"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_user",
        sa.Column(
            "auto_watch_threads_i_create",
            sa.Boolean(),
            nullable=False,
            server_default="1",
        ),
    )
    op.add_column(
        "rb_user",
        sa.Column(
            "auto_watch_threads_i_participate",
            sa.Boolean(),
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "rb_user",
        sa.Column(
            "default_node_watcher_frequency",
            sa.Enum("immediately", "daily", "weekly", name="frequency"),
            nullable=False,
            server_default="daily",
        ),
    )


def downgrade():
    op.drop_column("rb_user", "default_node_watcher_frequency")
    op.drop_column("rb_user", "auto_watch_threads_i_participate")
    op.drop_column("rb_user", "auto_watch_threads_i_create")



"""namespace group conversations

Revision ID: 6db39621f46c
Revises: f7b42b2024fb
Create Date: 2019-03-22 09:40:30.185446

"""

# revision identifiers, used by Alembic.
revision = '6db39621f46c'
down_revision = 'f7b42b2024fb'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('rb_namespace', sa.Column('group_conversations', sa.Boolean(), nullable=True))


def downgrade():
    pass



"""add subscription_type to Namespace

Revision ID: 3ee195d142fb
Revises: 771ef7c868de
Create Date: 2019-02-16 14:23:03.175524

"""

# revision identifiers, used by Alembic.
revision = '3ee195d142fb'
down_revision = '771ef7c868de'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'rb_namespace',
        sa.Column(
            'subscription_type',
            sa.Enum('development', 'trial', 'production', name='subscription_type'),
            nullable=False,
            server_default='development'
        )
    )


def downgrade():
    pass



"""root-stats

Revision ID: 503a2cef2378
Revises: aa2dfb4decda
Create Date: 2018-03-25 22:20:39.436208

"""

# revision identifiers, used by Alembic.
revision = "503a2cef2378"
down_revision = "aa2dfb4decda"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


from sqlalchemy_utils import UUIDType as TempUUIDType

UUIDType = TempUUIDType(binary=False)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "rb_root_cache",
        sa.Column("id", UUIDType, nullable=False),
        sa.Column("root_id", UUIDType, nullable=True),
        sa.Column("valid", sa.Boolean(), nullable=True),
        sa.Column("updated_timestamp", sa.BigInteger(), nullable=False),
        sa.Column("json_stats", sa.UnicodeText(), nullable=True),
        sa.ForeignKeyConstraint(["root_id"], ["rb_node.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_rb_root_cache_id"), "rb_root_cache", ["id"], unique=False)
    op.create_index(
        op.f("ix_rb_root_cache_root_id"), "rb_root_cache", ["root_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_rb_root_cache_root_id"), table_name="rb_root_cache")
    op.drop_index(op.f("ix_rb_root_cache_id"), table_name="rb_root_cache")
    op.drop_table("rb_root_cache")
    # ### end Alembic commands ###



from pyramid.paster import bootstrap, setup_logging

from ..lib.notify import deliver_scheduled_notifications

from . import base_parser


def get_arg_parser():
    parser = base_parser("Send Node Notification Digests.")
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env:
        deliver_scheduled_notifications(env["request"])



from .meta import Base, RBase, UUIDType, foreign_key, get_object_by_id, now_timestamp

from sqlalchemy import Column, BigInteger, Boolean, Unicode, UnicodeText, Enum

from sqlalchemy.orm import relationship

import uuid

from remarkbox.lib import (
    timestamp_to_datetime,
    timestamp_to_ago_string,
    timestamp_to_date_string,
)


class NodeEventNotification(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    # the Node this event is for.
    node_event_id = Column(UUIDType, foreign_key("NodeEvent", "id"), index=True)
    # the watcher which caused this notification to exist.
    watcher_id = Column(UUIDType, foreign_key("Watcher", "id"), index=True)
    # the user to notify.
    user_id = Column(UUIDType, foreign_key("User", "id"), index=True)
    # the type of action that was performed on Node.
    created_timestamp = Column(BigInteger, nullable=False)
    # the timestamp when the notification was last updated.
    updated_timestamp = Column(BigInteger, nullable=False)
    # was the notification sent yet?
    sent = Column(Boolean, default=False)
    # was the notification read yet?
    # not really used for email, but could be useful for
    # push notifications, web notifications, or browser notifications
    read = Column(Boolean, default=False)

    def __init__(self, node_event, watcher):
        self.id = uuid.uuid1()
        self.node_event = node_event
        self.watcher = watcher
        self.user = watcher.user
        self.created_timestamp = now_timestamp()
        self.updated_timestamp = now_timestamp()

    # 1-to-1 relationships.
    node_event = relationship(argument="NodeEvent", uselist=False, lazy="joined")
    watcher = relationship(argument="Watcher", uselist=False, lazy="joined")
    user = relationship(argument="User", uselist=False, lazy="joined")

    @property
    def frequency(self):
        return self.watcher.notification_frequency

    @property
    def method(self):
        return self.watcher.notification_method

    # TODO: take into account user's timezone?
    @property
    def human_created_timestamp(self):
        return timestamp_to_ago_string(self.created_timestamp)

    @property
    def created_date(self):
        return timestamp_to_date_string(self.created_timestamp)

    @property
    def created_datetime(self):
        return timestamp_to_datetime(self.created_timestamp)

    @property
    def human_changed_timestamp(self):
        return timestamp_to_ago_string(self.changed_timestamp)

    @property
    def changed_date(self):
        return timestamp_to_date_string(self.changed_timestamp)

    @property
    def changed_datetime(self):
        return timestamp_to_datetime(self.changed_timestamp)


def get_node_event_notification_by_id(dbsession, node_event_notification_id):
    """Try to get NodeEventNotification object by id or return None."""
    return get_object_by_id(dbsession, node_event_notfication_id, NodeEventNotification)


def get_topsecret_notifications(dbsession, limit=100, offset=0):
    """Return all NodeEventNotifications."""
    return (
        dbsession.query(NodeEventNotification)
        .order_by(
            NodeEventNotification.created_timestamp.desc(),
            NodeEventNotification.user_id,
        )
        .limit(limit)
        .offset(offset)
    )



"""new namespace columns 2

Revision ID: a738d6859f18
Revises: 98ced5f6ff36
Create Date: 2017-11-06 00:05:09.284352

"""

# revision identifiers, used by Alembic.
revision = "a738d6859f18"
down_revision = "98ced5f6ff36"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_namespace",
        sa.Column("google_analytics_id", sa.Unicode(length=18), nullable=True),
    )
    op.add_column(
        "rb_namespace",
        sa.Column("no_nodes_text", sa.Unicode(length=256), nullable=True),
    )
    op.add_column(
        "rb_namespace",
        sa.Column("owner_request_timestamp", sa.BigInteger(), nullable=True),
    )


def downgrade():
    op.drop_column("rb_namespace", "owner_request_timestamp")
    op.drop_column("rb_namespace", "no_nodes_text")
    op.drop_column("rb_namespace", "google_analytics_id")



from pyramid.paster import bootstrap, setup_logging

from ..models import (
    get_user_by_email,
    get_namespace_by_name,
    get_namespace_request,
    is_user_name_valid,
    is_user_name_available,
)

from . import base_parser

from six import u as unicode


def get_arg_parser():
    parser = base_parser("Modify a user.")
    parser.add_argument("-i", "--info", action="store_true", default=False)
    parser.add_argument("email", type=unicode)
    parser.add_argument("--namespace-remove", default=None, type=unicode)
    parser.add_argument("--namespace-add", default=None, type=unicode)
    parser.add_argument(
        "--namespace-role", default=None, type=unicode, choices=[None, "owner"]
    )
    parser.add_argument("--change-user-name", type=unicode)
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        user = get_user_by_email(request.dbsession, args.email)

        if user is None:
            print("No such user found by email: {}.".format(args.email))
            exit()

        if args.info:
            from collections import OrderedDict

            user_dict = OrderedDict(sorted(user.__dict__.items(), key=lambda t: t[0]))
            del user_dict["password"]
            del user_dict["email_id"]
            for key, value in user_dict.items():
                print("{}: {}".format(key, value))

            print("namespaces:")
            for namespace in user.namespaces:
                print(" - {}".format(namespace.name))
            exit()

        if args.namespace_add:
            namespace_to_add = get_namespace_by_name(
                request.dbsession, args.namespace_add
            )
            if namespace_to_add:
                namespace_to_add.set_role_for_user(user, role=args.namespace_role)
        if args.namespace_remove:
            namespace_to_remove = get_namespace_by_name(
                request.dbsession, args.namespace_remove
            )
            if namespace_to_remove:
                namespace_request_to_remove = get_namespace_request(
                    request.dbsession, user, namespace_to_remove
                )
                if namespace_request_to_remove:
                    request.dbsession.delete(namespace_request_to_remove)
                namespace_to_remove.clear_owner_request_timestamp()
                namespace_to_remove.remove_user(user)

        if args.change_user_name:
            if is_user_name_valid(args.change_user_name) and is_user_name_available(
                request.dbsession, args.change_user_name
            ):
                user.name = args.change_user_name

        request.dbsession.add(user)
        request.dbsession.flush()



from pyramid.paster import bootstrap, setup_logging

from remarkbox.models import recompute_depth_on_all_node_objects

from . import base_parser


def main():
    parser = base_parser(
        "Recompute nested node.depth in relation their thread."
    )
    parser.add_argument("--chunk", type=int, default=500,)
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        # Run node.recompute_depth() on all Node objects.
        # Flush all changes stored in the sqlalchemy session to database.
        recompute_depth_on_all_node_objects(request.dbsession, args.chunk)



"""pay what you can table

Revision ID: b132ec968f0c
Revises: 184a349cd589
Create Date: 2021-01-31 09:47:47.220938

"""

# revision identifiers, used by Alembic.
revision = 'b132ec968f0c'
down_revision = '184a349cd589'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from sqlalchemy_utils import UUIDType as TempUUIDType
UUIDType = TempUUIDType(binary=False)


def upgrade():
    op.create_table('rb_pay_what_you_can',
        sa.Column('user_id', UUIDType, nullable=False),
        sa.Column('frequency', sa.Enum('once', 'yearly', name='frequency'), nullable=False),
        sa.Column('amount', sa.BigInteger(), nullable=False),
        sa.Column('contributions', sa.BigInteger(), nullable=False),
        sa.Column('created_timestamp', sa.BigInteger(), nullable=False),
        sa.Column('updated_timestamp', sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], ['rb_user.id'], ),
        sa.PrimaryKeyConstraint('user_id')
    )
    op.create_index(op.f('ix_rb_pay_what_you_can_user_id'), 'rb_pay_what_you_can', ['user_id'], unique=False)



from hashlib import md5

from ago import human

from datetime import datetime


def timestamp_to_datetime(timestamp):
    """Accepts a timestamp and returns a date string"""
    return datetime.fromtimestamp(timestamp / 1000.0)


def timestamp_to_date_string(timestamp):
    return timestamp_to_datetime(timestamp).strftime("%b %d, %Y %I:%M %P")


def timestamp_to_ago_string(timestamp):
    """Accepts a timestamp and returns a human readable string"""
    return human(timestamp_to_datetime(timestamp), 2, abbreviate=True)


def gravatar_client(email, size=48, rating="g", default="retro", force_default=False):
    """Return a gravatar API URI."""
    query_params = ["s={}".format(size), "r={}".format(rating), "d={}".format(default)]
    if force_default:
        query_params.append("f=y")
    uri = "https://secure.gravatar.com/avatar/{0}?{1}"

    try:
        # Python 2.
        email_md5 = md5(email).hexdigest()
    except TypeError:
        # Python 3.
        email_md5 = md5(email.encode("utf-8")).hexdigest()

    return uri.format(email_md5, "&".join(query_params))



from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound

from remarkbox.views import (
    get_referer_or_home,
    get_join_or_log_in_route_uri,
    user_required,
)

from remarkbox.models import get_oauth_by_id


def get_namespace_settings_route(request, anchor=None):
    return request.route_url(
        "basic-namespace-settings", namespace=request.namespace.name, _anchor=anchor
    )


@view_config(route_name="oauth-slack-delete")
@user_required()
def oauth_slack_delete(request):
    oauth_id = request.params.get("oauth-id", "")
    oauth_record = get_oauth_by_id(request.dbsession, oauth_id)

    if oauth_record:
        request.dbsession.delete(oauth_record)
        request.session.flash(
            ("You deleted that Remarkbox side of that Slack integration.", "success")
        )
        request.session.flash(
            (
                "Please remember to delete the Slack Team side of the integration.",
                "info",
            )
        )
        request.dbsession.flush()
    else:
        request.session.flash(("Invalid OAuth Id.", "error"))

    return HTTPFound(get_namespace_settings_route(request, "notifications"))


@view_config(route_name="oauth-slack")
@user_required()
def oauth_slack(request):

    if not request.user in request.namespace.owners:
        request.session.flash(("You do not own that Namespace.", "error"))
        return HTTPFound(get_join_or_log_in_route_uri(request))

    oauth_error = request.params.get("error", "")
    oauth_code = request.params.get("code", "")

    if oauth_error:
        if oauth_error == "access_denied":
            request.session.flash(
                ("You declined to grant Remarkbox access to Slack", "info")
            )
        else:
            request.session.flash((oauth_error, "error"))
        return HTTPFound(get_namespace_settings_route(request, "notifications"))

    from slacker import Slacker

    # when doing the oauth dance, the first time
    # we connect we don't need a token.
    slack = Slacker("")

    # Request the auth tokens from Slack
    oauth_response = slack.oauth.access(
        client_id=request.app.get("slack.public"),
        client_secret=request.app.get("slack.secret"),
        code=oauth_code,
        redirect_uri=request.route_url(
            "oauth-slack", _query={"namespace": request.namespace.name}
        ),
    )

    if oauth_response.successful:

        #access_token = oauth_response.body["bot"]["bot_access_token"]
        access_token = oauth_response.body["access_token"]
        request.namespace.add_oauth_record(
            user=request.user,
            service="slack",
            token=access_token,
            data=oauth_response.body
        )
        request.dbsession.add(request.namespace)
        request.dbsession.flush()
        request.session.flash(
            (
                "Success, you integrated Remarkbox with <b>{}</b> (Slack Team)".format(
                    oauth_response.body["team_name"]
                ),
                "success",
            )
        )
        request.session.flash(
            (
                "Please create a channel in <b>{}</b> called <b>#remarks</b>".format(
                    oauth_response.body["team_name"]
                ),
                "info",
            )
        )

        # send a test message to #remark channel!
        msg1 = "*Success!* {} integrated this Slack Team with a Remarkbox Namespace (`{}`)".format(
            request.user.name,
            request.namespace.name,
        )
        msg2 = "Please create the `#remarks` channel."
        slack = Slacker(access_token)
        slack.chat.post_message(
            "#general",
            msg1,
        )
        slack.chat.post_message(
            "#general",
            msg2,
        )

    return HTTPFound(get_namespace_settings_route(request, "notifications"))



"""autogenerated indicies.

Revision ID: 04e52337756e
Revises: 6a440d3921d2
Create Date: 2017-07-03 13:45:32.742934

"""

# revision identifiers, used by Alembic.
revision = "04e52337756e"
down_revision = "6a440d3921d2"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f("ix_Node_id"), "Node", ["id"], unique=False)
    op.create_index(op.f("ix_Node_root_id"), "Node", ["root_id"], unique=False)
    op.create_index(op.f("ix_Uri_data"), "Uri", ["data"], unique=False)
    op.create_index(op.f("ix_Uri_id"), "Uri", ["id"], unique=False)
    op.create_index(op.f("ix_User_id"), "User", ["id"], unique=False)
    op.create_index(op.f("ix_Vote_id"), "Vote", ["id"], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_Vote_id"), table_name="Vote")
    op.drop_index(op.f("ix_User_id"), table_name="User")
    op.drop_index(op.f("ix_Uri_id"), table_name="Uri")
    op.drop_index(op.f("ix_Uri_data"), table_name="Uri")
    op.drop_index(op.f("ix_Node_root_id"), table_name="Node")
    op.drop_index(op.f("ix_Node_id"), table_name="Node")
    ### end Alembic commands ###



"""new namespace column 3

Revision ID: 43c1e5a99d4e
Revises: a738d6859f18
Create Date: 2017-11-11 10:35:40.930389

"""

# revision identifiers, used by Alembic.
revision = "43c1e5a99d4e"
down_revision = "a738d6859f18"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_namespace",
        sa.Column("placeholder_text", sa.Unicode(length=256), nullable=True),
    )


def downgrade():
    op.drop_column("rb_namespace", "placeholder_text")



from .meta import NODE_EVENT_ACTIONS
from .meta import Base, RBase, UUIDType, foreign_key, get_object_by_id, now_timestamp

from sqlalchemy import Column, BigInteger, Unicode, UnicodeText, Enum

from sqlalchemy.orm import relationship, backref

import uuid


class NodeEvent(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    # the Node this event is for.
    node_id = Column(UUIDType, foreign_key("Node", "id"), index=True)
    # the User who performed the event action on the Node.
    user_id = Column(UUIDType, foreign_key("User", "id"), index=True)
    # the type of action that was performed on Node.
    action = Column(Enum(*NODE_EVENT_ACTIONS, name="action"), nullable=False)
    # the timestamp when the event was created.
    created_timestamp = Column(BigInteger, nullable=False)

    def __init__(self, user, node, action):
        self.id = uuid.uuid1()
        self.user = user
        self.node = node
        self.action = action
        self.created_timestamp = now_timestamp()

    # 1-to-1 relationships.
    node = relationship(argument="Node", uselist=False, lazy="joined")
    user = relationship(
        argument="User",
        uselist=False,
        lazy="joined",
        backref=backref("events", cascade="save-update, merge, delete"),
    )


def get_node_event_by_id(dbsession, node_event_id):
    """Try to get NodeEvent object by id or return None."""
    return get_object_by_id(dbsession, node_event_id, NodeEvent)



import json

from pyramid.paster import bootstrap, setup_logging

from ..models import (
    User,
    UserSurrogate,
    generate_user_name,
    get_user_by_email,
    get_user_surrogate_by_name,
    get_or_create_node_by_uri,
    is_user_name_valid,
    is_user_name_available,
)

from . import base_parser

try:
    unicode("")
except:
    from six import u as unicode


def generate_password(size=32):
    """Return a system generated password"""
    from random import choice
    letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    digits = "0123456789"
    pool = letters + digits
    return "".join([choice(pool) for i in range(size)])


def get_arg_parser():
    parser = base_parser("import comments from JSON")
    # parser.add_argument('--dry-run', action='store_true', default=False)
    parser.add_argument('--user-surrogates', action='store_true', default=False)
    parser.add_argument("--group")
    parser.add_argument("json_path", help="path to JSON file")
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    # group = generate_password(3)
    group = ""
    if args.group:
        group = args.group

    with open(args.json_path) as f:
        pages = json.load(f)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        users = {}
        user_surrogates = {}
        nodes = {}

        for slug, page_data in pages.items():

            comments = page_data["comments"]

            if len(comments) == 0:
                # skipping because this page has no comments.
                continue

            uri = page_data["link"]

            print("creating page for {}".format(uri))
            root = get_or_create_node_by_uri(request.dbsession, uri)
            request.dbsession.add(root)
            request.dbsession.flush()

            for comment in comments:

                email = comment["email"]

                user = None
                user_surrogate = None

                if not email:
                    if args.user_surrogates:
                        name = comment["author"]
                        if name not in user_surrogates:
                            user_surrogates[name] = UserSurrogate(
                                name,
                                root.namespace,
                            )
                        user_surrogate = user_surrogates[name]
                        request.dbsession.add(user_surrogate)
                        request.dbsession.flush()
                    else:
                        # an email may not be empty or None.
                        continue

                elif email in users:
                    user = users[email]
                else:
                    user = get_user_by_email(request.dbsession, email)
                    if user:
                        users[email] = user
                        print("found user for {} ({})".format(email, user.name))
                    else:
                        desired_name = (
                            comment["author"].replace(" ", "-").replace("_", "-")
                        )
                        if group:
                            desired_name = desired_name + "-" + group
                        if not is_user_name_valid(
                            desired_name
                        ) or not is_user_name_available(
                            request.dbsession, desired_name
                        ):
                            desired_name = generate_user_name(
                                request.dbsession, desired_name
                            )
                        user = User(email)
                        user.name = unicode(desired_name)
                        users[email] = user
                        print("created user for {} ({})".format(email, desired_name))

                        request.dbsession.add(user)
                        request.dbsession.flush()


                node = root.new_child()
                node.set_data(unicode(comment["content"]))
                node.created = int(comment["timestamp"]) * 1000
                node.changed = int(comment["timestamp"]) * 1000

                if user is not None:
                    node.user = user
                elif user_surrogate is not None:
                    node.user_surrogate = user_surrogate

                node.verified = True
                node.ip_address = comment.get("author_ip", None)
                nodes[comment["id"]] = node

                request.dbsession.add(nodes[comment["id"]])
                request.dbsession.flush()

            for comment in comments:
                if comment["parent_id"]:
                    print(
                        "id {}, parent_id {}".format(
                            comment["id"], comment["parent_id"]
                        )
                    )
                    if nodes.get(comment["parent_id"]):
                        nodes[comment["id"]].parent_id = nodes[comment["parent_id"]].id
                        request.dbsession.add(nodes[comment["id"]])
                        request.dbsession.flush()
                    else:
                        print(
                            "parent_id ({}) is not found, skipping".format(
                                comment["parent_id"]
                            )
                        )



"""add mathjax column to rb_namespace

Revision ID: f5f88293428a
Revises: ba8e83a3b8fd
Create Date: 2017-10-14 23:13:46.584415

"""

# revision identifiers, used by Alembic.
revision = "f5f88293428a"
down_revision = "ba8e83a3b8fd"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_namespace", sa.Column("mathjax", sa.Boolean(), server_default=sa.false())
    )


def downgrade():
    op.drop_column("rb_namespace", "mathjax")



"""create graph_depth column on node model.

Revision ID: 76deff88cfd6
Revises: d1a0ab4f617f
Create Date: 2020-02-23 20:51:49.889183

"""

# revision identifiers, used by Alembic.
revision = '76deff88cfd6'
down_revision = 'd1a0ab4f617f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'rb_node',
        sa.Column(
            'graph_depth',
            sa.Integer(),
            nullable=False,
            server_default="-1",
        )
    )

def downgrade():
    op.drop_column('rb_node', 'graph_depth')



from collections import defaultdict

from pyramid.paster import bootstrap, setup_logging

from ..models import get_namespace_by_name, get_nodes_who_share_root

from . import base_parser


def get_arg_parser():
    parser = base_parser("Given a namespace find dupe uri roots and merge them.")
    parser.add_argument("namespace")
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    uris = defaultdict(list)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        namespace = get_namespace_by_name(request.dbsession, args.namespace)
        if namespace is not None:

            for root in namespace.roots:
                if root.uri is None:
                    continue
                uris[root.uri.data].append(root)

            for uri, roots in uris.items():
                if len(roots) == 2:
                    for child in roots[0].children:
                        child.parent_id = roots[1].id
                        request.dbsession.add(child)
                    nodes = get_nodes_who_share_root(request.dbsession, roots[0])
                    for node in nodes:
                        node.root_id = roots[1].id
                        request.dbsession.add(node)
                    request.dbsession.flush()
                    request.dbsession.delete(roots[0])
                    request.dbsession.delete(roots[0].uri)
                    request.dbsession.flush()

        else:
            print("no such namespace ({})".format(args.namespace))



from __future__ import print_function

from pyramid.paster import bootstrap, setup_logging

from ..models import get_or_create_namespace, get_namespace_by_name, get_nodes_who_share_roots

from ..models.meta import SUBSCRIPTION_TYPES

from . import base_parser


toggle = lambda x: not x


def get_arg_parser():
    parser = base_parser("Modify a Namespace.")
    parser.add_argument(
        "--toggle-hide-unless-approved", action="store_true", default=False
    )
    parser.add_argument("--create", action="store_true", default=False)
    parser.add_argument("--delete", action="store_true", default=False)
    parser.add_argument("--delete-children-nodes", action="store_true", default=False, help="this is useful if you mess up an import.")
    parser.add_argument("--info", action="store_true", default=False)
    parser.add_argument("--subscription-type", choices=SUBSCRIPTION_TYPES)
    parser.add_argument("namespace")
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        sp = tm.savepoint()
        request = env["request"]

        if args.create:
            namespace = get_or_create_namespace(request.dbsession, args.namespace)
            request.dbsession.add(namespace)
            request.dbsession.flush()
        else:
            namespace = get_namespace_by_name(request.dbsession, args.namespace)

        if not namespace:
            print("Namespace '{}' not found.".format(args.namespace))
            exit()

        if args.info:
            from collections import OrderedDict

            namespace_dict = OrderedDict(
                sorted(namespace.__dict__.items(), key=lambda t: t[0])
            )
            for key, value in namespace_dict.items():
                print("{}: ".format(key), end="")
                print(value)

            print("owners:")
            for user in namespace.owners:
                print("  {}: {}".format(user.name, user.email))

            print("moderators:")
            for user in namespace.moderators:
                print("  {}: {}".format(user.name, user.email))

            exit()

        elif args.subscription_type:
            namespace.subscription_type = args.subscription_type
            request.dbsession.add(namespace)
            request.dbsession.flush()

        elif args.toggle_hide_unless_approved:
            namespace.hide_unless_approved = toggle(namespace.hide_unless_approved)
            request.dbsession.add(namespace)
            request.dbsession.flush()

        elif args.delete_children_nodes:
            nodes = get_nodes_who_share_roots(request.dbsession, namespace.roots)

            # delete all children nodes.
            for node in nodes:
                if node.is_root:
                    # Skip root nodes to keep Uri and Watchers relations.
                    continue
                request.dbsession.delete(node)

            if (
                raw_input("*** DANGER: Really delete Nodes forever? [yes, no]: ")
                == "yes"
            ):
                request.dbsession.flush()
                print("Flushed transaction to database, the Nodes were completely destroyed!")
            else:
                sp.rollback()
                print("Rolled back transaction, sheepishly refused to delete anything!")

        elif args.delete:

            nodes = get_nodes_who_share_roots(request.dbsession, namespace.roots)

            # delete all nodes.
            for node in nodes:
                if node.events:
                    # delete all related NodeEvents.
                    for event in node.events:
                        request.dbsession.delete(event)

                # delete node.
                request.dbsession.delete(node)

            for root in namespace.roots:
                if root.uri:
                    # delete related Uri.
                    request.dbsession.delete(root.uri)
                if root.cache:
                    # delete related NodeCache.
                    request.dbsession.delete(root.cache)
                if root.watchers:
                    # delete all related NodeEventWatchers.
                    for watcher in root.watchers:
                        # delete all unsent NodeEventNotifications.
                        if watcher.unsent_notifications():
                            for notification in watcher.unsent_notifications():
                                request.dbsession.delete(notification)
                        request.dbsession.delete(watcher)

            if namespace.watchers:
               for watcher in namespace.watchers:
                   # delete all unsent NodeEventNotifications.
                   if watcher.unsent_notifications():
                       for notification in watcher.unsent_notifications():
                           request.dbsession.delete(notification)
                   request.dbsession.delete(watcher)

            # disable all related NamespaceUser objects.
            if namespace.namespace_users:
                for nsu in namespace.namespace_users:
                    # delete related NamespaceUser.
                   request.dbsession.delete(nsu)

            # delete all related OauthRecords.
            if namespace.oauth_records:
                for oauth_record in namespace.oauth_records:
                    request.dbsession.delete(oauth_record)

            # finally delete Namespace.
            request.dbsession.delete(namespace)

            if (
                raw_input("*** DANGER: Delete Namespace '{}' ({}) forever? [yes, no]: ".format(namespace.name, namespace.id))
                == "yes"
            ):
                request.dbsession.flush()
                print("Flushed transaction to database, the Namespace was completely destroyed!")
            else:
                sp.rollback()
                print("Rolled back transaction, sheepishly refused to delete anything!")



from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound, HTTPBadRequest

from remarkbox.views import user_required

from remarkbox.lib.mail import send_operator_email

from remarkbox.models import PayWhatYouCan


@view_config(route_name="billing", renderer="billing.j2")
@user_required(
    flash_msg="Thank you for helping us out, Please verify your email in the form below!",
    flash_level="info",
    return_to_route_name="billing"
)
def billing(request):
    return {"the_title": "Payment Preferences"}


@view_config(route_name="add-card")
@user_required()
def add_card(request):
    # try to get the return_to uri from posted parameters.
    return_to = request.params.get("return-to", "/billing")
    try:
        source = request.stripe_customer.sources.create(
            source=request.params.get("stripeToken")
        )
        request.stripe_customer.default_source = source
        request.session.flash(("You saved a new card.", "success"))
        # TODO: this is very temporary just as a stop gap for me to stay on top of new customers.
        send_operator_email(
            request,
            "A user (hopefully a new one!) added a new card to their Stripe customer account. Log into Stripe and follow up to close the sale!",
        )
    except request.stripe.error.CardError as e:
        body = e.json_body
        err = body.get("error", {})
        request.session.flash((err.get("message"), "error"))
    request.stripe_customer.save()
    return HTTPFound(return_to)


@view_config(route_name="confirm-update-card", renderer="update-card.j2")
@user_required()
def confirm_update_card(request):
    card_id = request.matchdict.get("card_id")
    card = request.stripe_customer.sources.retrieve(card_id)
    action = request.matchdict.get("action")
    action_human = action.replace("-", " ")
    button_class = "red-button" if action == "delete-card" else "blue-button"
    return {
        "card": card,
        "card_id": card_id,
        "action": action,
        "action_human": action_human,
        "button_class": button_class,
        "the_title": action_human.title(),
    }


@view_config(route_name="update-card")
@user_required()
def update_card(request):
    action = request.params.get("action", None)
    card_id = request.params.get("card_id")
    if action not in ["make-card-active", "delete-card"]:
        return HTTPBadRequest
    if "delete-card" == action:
        request.stripe_customer.sources.retrieve(card_id).delete()
        request.session.flash(("You deleted that card.", "success"))
    if "make-card-active" == action:
        request.stripe_customer.default_source = card_id
        request.session.flash(("You set the active card.", "success"))
    request.stripe_customer.save()
    return HTTPFound("/billing")


@view_config(route_name="pay-what-you-can")
@user_required()
def pay_what_you_can(request):
    frequency = request.params.get("frequency", None)
    amount = request.params.get("amount", None)
    if frequency is None or amount is None:
        request.session.flash(("You must put a value for both frequency and amount.", "error"))
    elif request.user.pay_what_you_can:
        request.user.pay_what_you_can.update(frequency, amount)
        request.session.flash(("You updated your contribution preferences.", "success"))
    else:
        request.user.pay_what_you_can = PayWhatYouCan(request.user, frequency, amount)
        request.session.flash(("You updated your contribution preferences.", "success"))
    return HTTPFound("/billing")



"""new field for namespace request cache

Revision ID: 14a6a35940c7
Revises: 8f76e372a953
Create Date: 2024-10-06 09:17:13.978089

"""

# revision identifiers, used by Alembic.
revision = '14a6a35940c7'
down_revision = '8f76e372a953'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa



from sqlalchemy_utils import UUIDType as TempUUIDType
UUIDType = TempUUIDType(binary=False)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('rb_namespace_request', sa.Column('last_scrape_timestamp', sa.BigInteger(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('rb_namespace_request', 'last_scrape_timestamp')
    # ### end Alembic commands ###



"""Node disabled_timestamp

Revision ID: 2e00b92495e4
Revises: 7a3a4755b4a8
Create Date: 2017-08-13 15:58:33.169624

"""

# revision identifiers, used by Alembic.
revision = "2e00b92495e4"
down_revision = "7a3a4755b4a8"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import Node, now_timestamp


def upgrade():
    # add the new column.
    op.add_column(
        "Node", sa.Column("disabled_timestamp", sa.BigInteger(), nullable=True)
    )

    bind = op.get_bind()
    DBSession = sa.orm.Session(bind=bind)

    # get all disabled nodes.
    nodes = DBSession.query(Node).filter(Node.disabled == True)

    for node in nodes:
        node.disabled_timestamp = now_timestamp()
        DBSession.add(node)

    DBSession.flush()
    DBSession.commit()


def downgrade():
    op.drop_column("Node", "disabled_timestamp")



from .meta import Base, RBase, UUIDType, foreign_key, get_object_by_id, now_timestamp

from sqlalchemy import Column, BigInteger, Unicode, UnicodeText

from sqlalchemy.orm import relationship

import uuid

import json

try:
    unicode("")
except:
    from six import u as unicode


class Oauth(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    # the related user this oauth record belongs to.
    user_id = Column(UUIDType, foreign_key("User", "id"), index=True)
    # the related namespace this oauth record belongs to.
    namespace_id = Column(UUIDType, foreign_key("Namespace", "id"), index=True)
    # the name of the service (type of oauth record), for example slack.
    service = Column(Unicode(16), nullable=False)
    # the token provided after oauth dance used to communicate with 3rd party.
    token = Column(Unicode(256), nullable=False)
    # optional json_data without defined schema.
    json_data = Column(UnicodeText, default=None)
    # timestamp when Namespace owner authorized consent to connect with 3rd party oauth.
    created_timestamp = Column(BigInteger, nullable=False)
    # the last time this record was modified.
    updated_timestamp = Column(BigInteger, nullable=False)

    def __init__(self, user, namespace, service, token, data=None):
        self.id = uuid.uuid1()
        self.user = user
        self.namespace = namespace
        self.service = unicode(service)
        self.token = unicode(token)
        self.created_timestamp = now_timestamp()
        self.data = data

    # 1-to-1 relationships.
    user = relationship(argument="User", uselist=False, lazy="joined")
    namespace = relationship(argument="Namespace", uselist=False, lazy="joined")

    @property
    def data(self):
        if self.json_data:
            return json.loads(self.json_data)
        return None

    @data.setter
    def data(self, new_data):
        if new_data:
            self.json_data = json.dumps(new_data)
        else:
            self.json_data = None
        self.updated_timestamp = now_timestamp()


def get_oauth_by_id(dbsession, oauth_id):
    """Try to get Oauth object by id or return None"""
    return get_object_by_id(dbsession, oauth_id, Oauth)



from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound

from pyramid.response import Response

from remarkbox.models.meta import NOTIFICATION_FREQUENCIES

from remarkbox.models import (
    get_or_create_namespace,
    get_or_create_namespace_request,
    get_namespace_request_by_id,
    is_user_name_available,
    is_user_name_valid,
)

from remarkbox.views import get_referer_or_home, user_required, reject_stand_alone

from string import digits

try:
    # Python 2.
    from string import lowercase
except ImportError:
    # Python 3.
    from string import ascii_lowercase as lowercase


domain_name_set = set(lowercase + digits + ".-")


def checkbox_to_bool(checkbox):
    return True if checkbox == "on" else False


@view_config(route_name="basic-namespace-settings", renderer="namespace-settings.j2")
@view_config(route_name="embed-namespace-settings", renderer="namespace-settings.j2")
@user_required()
def namespace_settings(request):

    if not request.user in request.namespace.owners:
        request.session.flash(("You do not own that Namespace.", "error"))
        return HTTPFound(get_referer_or_home(request))

    # temporarily set attribute protection to False
    # so that an unpaid Namespace still displays the set values.
    request.namespace.memoized_attr_protection = False

    if request.method == "POST":
        p = request.params
        stylesheet_embed = p.get("stylesheet-embed", request.namespace.stylesheet_embed)
        stylesheet_embed_uri = p.get(
            "stylesheet-embed-uri", request.namespace.stylesheet_embed_uri
        )
        description = p.get("description", request.namespace.description)
        no_nodes_text = p.get("no-nodes-text", request.namespace.no_nodes_text)
        placeholder_text = p.get("placeholder-text", request.namespace.placeholder_text)
        google_analytics_id = p.get(
            "google-analytics-id", request.namespace.google_analytics_id
        )

        hide_unless_approved_checkbox = p.get("hide-unless-approved-checkbox", "off")
        link_protection_checkbox = p.get("link-protection-checkbox", "off")
        reverse_order_checkbox = p.get("reverse-order-checkbox", "off")
        group_conversations_checkbox = p.get("group-conversations-checkbox", "off")
        mathjax_checkbox = p.get("mathjax-checkbox", "off")
        ignore_query_string_checkbox = p.get("ignore-query-string-checkbox", "off")
        hide_powered_by_checkbox = p.get("hide-powered-by-checkbox", "off")

        hide_unless_approved = checkbox_to_bool(hide_unless_approved_checkbox)
        link_protection = checkbox_to_bool(link_protection_checkbox)
        reverse_order = checkbox_to_bool(reverse_order_checkbox)
        group_conversations = checkbox_to_bool(group_conversations_checkbox)
        mathjax = checkbox_to_bool(mathjax_checkbox)
        ignore_query_string = checkbox_to_bool(ignore_query_string_checkbox)
        hide_powered_by = checkbox_to_bool(hide_powered_by_checkbox)

        if stylesheet_embed != request.namespace.stylesheet_embed and (
            stylesheet_embed or request.namespace.stylesheet_embed
        ):
            request.namespace.set_stylesheet_embed(stylesheet_embed)
            request.session.flash(("Success, you changed the stylesheet", "success"))

        if stylesheet_embed_uri != request.namespace.stylesheet_embed_uri and (
            stylesheet_embed_uri or request.namespace.stylesheet_embed_uri
        ):
            request.namespace.stylesheet_embed_uri = stylesheet_embed_uri
            request.session.flash(
                ("Success, you changed the stylesheet URI", "success")
            )

        if description != request.namespace.description and (
            description or request.namespace.description
        ):
            request.namespace.description = description
            request.session.flash(
                (
                    "Success, you changed the description for {}".format(
                        request.namespace.name
                    ),
                    "success",
                )
            )

        if no_nodes_text != request.namespace.no_nodes_text and (
            no_nodes_text or request.namespace.no_nodes_text
        ):
            request.namespace.no_nodes_text = no_nodes_text
            request.session.flash(
                (
                    "Success, you changed the text when there are no nodes in the thread.",
                    "success",
                )
            )

        if placeholder_text != request.namespace.placeholder_text and (
            placeholder_text or request.namespace.placeholder_text
        ):
            request.namespace.placeholder_text = placeholder_text
            request.session.flash(
                (
                    "Success, you changed the placeholder textarea text for new nodes.",
                    "success",
                )
            )

        if google_analytics_id != request.namespace.google_analytics_id and (
            google_analytics_id or request.namespace.google_analytics_id
        ):
            request.namespace.google_analytics_id = google_analytics_id
            request.session.flash(
                (
                    "Success, you changed the Google Analytics ID for stand alone mode.",
                    "success",
                )
            )

        if hide_unless_approved != request.namespace.hide_unless_approved:
            request.namespace.hide_unless_approved = hide_unless_approved
            request.session.flash(
                (
                    "You turned {} hide_unless_approved".format(
                        hide_unless_approved_checkbox
                    ),
                    "success",
                )
            )

        if link_protection != request.namespace.link_protection:
            request.namespace.link_protection = link_protection
            request.session.flash(
                (
                    "You turned {} link_protection".format(link_protection_checkbox),
                    "success",
                )
            )

        if reverse_order != request.namespace.reverse_order:
            request.namespace.reverse_order = reverse_order
            request.session.flash(
                (
                    "You turned {} reverse_order".format(reverse_order_checkbox),
                    "success",
                )
            )

        if group_conversations != request.namespace.group_conversations:
            request.namespace.group_conversations = group_conversations
            request.session.flash(
                (
                    "You turned {} group_conversations".format(
                        group_conversations_checkbox
                    ),
                    "success",
                )
            )

        if ignore_query_string != request.namespace.ignore_query_string:
            request.namespace.ignore_query_string = ignore_query_string
            request.session.flash(
                (
                    "You turned {} ignore_query_string".format(
                        ignore_query_string_checkbox
                    ),
                    "success",
                )
            )

        if hide_powered_by != request.namespace.hide_powered_by:
            request.namespace.hide_powered_by = hide_powered_by
            request.session.flash(
                (
                    "You turned {} hide_powered_by".format(hide_powered_by_checkbox),
                    "success",
                )
            )

        if mathjax != request.namespace.mathjax:
            request.namespace.mathjax = mathjax
            request.session.flash(
                ("You turned {} MathJax".format(mathjax_checkbox), "success")
            )

        request.dbsession.add(request.namespace)
        request.dbsession.flush()

    return {"the_title": "Namespace settings for {}".format(request.namespace.name)}


@view_config(route_name="embed-user-settings", renderer="user-settings.j2")
@view_config(route_name="basic-user-settings", renderer="user-settings.j2")
@user_required()
def user_settings(request):
    if request.method == "POST":
        _msg = "<b>{}</b> was turned <b>{}</b>"
        display_name = request.params.get("display-name", request.user.name)

        gravatar_checkbox = request.params.get("gravatar-checkbox", "off")
        auto_watch_create_checkbox = request.params.get(
            "auto-watch-threads-i-create-checkbox", "off"
        )
        auto_watch_participate_checkbox = request.params.get(
            "auto-watch-threads-i-participate-checkbox", "off"
        )

        gravatar = checkbox_to_bool(gravatar_checkbox)
        auto_watch_create = checkbox_to_bool(auto_watch_create_checkbox)
        auto_watch_participate = checkbox_to_bool(auto_watch_participate_checkbox)

        reply_watcher_frequency = request.params["reply-watcher-frequency"]
        default_node_watcher_frequency = request.params[
            "default-node-watcher-frequency"
        ]

        if display_name != request.user.name:
            if is_user_name_valid(display_name) == False:
                request.session.flash(
                    ("Sorry, display name must be alpha numeric", "error")
                )
            elif is_user_name_available(request.dbsession, display_name) == False:
                request.session.flash(
                    ('Sorry, "{}" is already in use'.format(display_name), "error")
                )
            else:
                request.user.name = display_name
                request.session.flash(
                    (
                        "Nice to meet you <b>{}</b>! (preferred display name saved!)".format(
                            request.user.name
                        ),
                        "success",
                    )
                )

        if gravatar != request.user.gravatar:
            request.user.gravatar = gravatar
            request.session.flash(
                (_msg.format("gravatar", gravatar_checkbox), "success")
            )

        """
        if auto_watch_create != request.user.auto_watch_threads_i_create:
            request.user.auto_watch_threads_i_create = auto_watch_create
            request.session.flash(
                (
                    _msg.format(
                        "auto_watch_threads_i_create"
                        auto_watch_create_checkbox,
                    ),
                    "success",
                )
            )

        if auto_watch_participate != request.user.auto_watch_threads_i_participate:
            request.user.auto_watch_threads_i_participate = auto_watch_participate
            request.session.flash(
                (
                    _msg.format(
                        "auto_watch_threads_i_participate",
                        auto_watch_participate_checkbox,
                    ),
                    "success",
                )
            )
        """

        if (
            default_node_watcher_frequency
            != request.user.default_node_watcher_frequency
        ):
            if default_node_watcher_frequency in NOTIFICATION_FREQUENCIES:
                request.user.default_node_watcher_frequency = (
                    default_node_watcher_frequency
                )
                request.session.flash(
                    (
                        "By default, when you <b>watch a thread</b>, we will notify you <b>{}</b>".format(
                            default_node_watcher_frequency
                        ),
                        "success",
                    )
                )
            else:
                request.session.flash(
                    ("Invalid Enum value for default_node_watcher_frequency.", "error")
                )

        if reply_watcher_frequency != request.user.reply_watcher.notification_frequency:
            if reply_watcher_frequency in NOTIFICATION_FREQUENCIES:
                request.user.reply_watcher.notification_frequency = (
                    reply_watcher_frequency
                )
                request.session.flash(
                    (
                        "When any of your comments get a reply, we will notify you <b>{}</b> via <b>{}</b>".format(
                            request.user.reply_watcher.notification_frequency,
                            request.user.reply_watcher.notification_method,
                        ),
                        "success",
                    )
                )
            else:
                request.session.flash(
                    ("Invalid Enum value for reply_watcher_frequency.", "error")
                )

        request.dbsession.add(request.user)
        request.dbsession.flush()

    return {"the_title": "{}'s settings".format(request.user.name)}


@view_config(route_name="setup", renderer="setup-namespace.j2")
@user_required(
    flash_msg="Verify your email to get started.",
    flash_level="info",
    return_to_route_name="setup",
)
@reject_stand_alone
def setup_namespace(request):
    if request.method == "POST":
        # domains are case insensitive (make lowercase).
        namespace_domain = request.params.get("namespace-domain", "").lower()
        if "." not in namespace_domain:
            request.session.flash(('A domain needs at least one "."', "error"))
        elif namespace_domain.startswith("-") or namespace_domain.startswith("."):
            request.session.flash(('A domain cannot start with "." or "-"', "error"))
        elif not set(namespace_domain).issubset(domain_name_set):
            request.session.flash(
                ('A domain has only letters, digits, "." and "-"', "error")
            )
        elif namespace_domain:
            namespace = get_or_create_namespace(request.dbsession, namespace_domain)
            namespace_request = get_or_create_namespace_request(
                request.dbsession, request.user, namespace
            )
            request.dbsession.add(namespace_request)
            request.dbsession.flush()
            request.session.flash(("Great work, on to Step 2!", "success"))
        else:
            request.session.flash(("Please enter a domain.", "error"))

    return {
        "namespaces": request.user.namespaces,
        "namespace_requests": request.user.namespace_owner_requests,
        "unverified_namespace_requests": request.user.unverified_namespace_owner_requests,
    }


@view_config(route_name="setup-namespace-request-cancel")
@user_required()
@reject_stand_alone
def namespace_request_cancel(request):
    if request.method == "POST":
        namespace_request_id = request.params.get("namespace-request-id")
        if namespace_request_id:
            namespace_request = get_namespace_request_by_id(
                request.dbsession, namespace_request_id
            )
            if namespace_request and namespace_request.user == request.user:
                request.dbsession.delete(namespace_request)
                request.dbsession.flush()
    return HTTPFound(get_referer_or_home(request))


@view_config(route_name="namespace-dump-to-json")
@user_required()
@reject_stand_alone
def namespace_dump_to_json(request):

    if not request.user in request.namespace.owners:
        request.session.flash(("You do not own that Namespace.", "error"))
        return HTTPFound(get_referer_or_home(request))

    from json import dumps

    response = Response(body=dumps(request.namespace.dict_dump))
    response.headerlist = []
    # allow Javascript from other domains to download this resource.
    response.headerlist.extend(
        (("Access-Control-Allow-Origin", "*"), ("Content-Type", "application/json"))
    )
    return response



WELCOME_1_TEXT = """
Hello!

Thanks for joining the discussion.

Hey there!

Here is the verification code you requested:

 \n{0}\n

Type code into the challenge input field.

What's next?

 * A random username was generated just for you!

Once signed in, you have the power to change your username as well as edit or delete past comments.
You may opt-in to get notified when people reply.

Talk to you soon!
"""

WELCOME_1_HTML = """
<!DOCTYPE html>
<html>
<head>
<title>{0}</title>
</head>
  <body>
    <h2>Hello!</h2>

    <p>
    Thanks for joining the discussion.
    </p>

    <p>
    Here is the verification code you requested:
    </p>

    <p><b>{1}</b></p>

    <p>
    This will verify your email and log you in.
    </p>

    <h3>What's next?</h3>

    <p>
    A random username was generated just for you!
    </p>

    <p>
    Once signed in, you have the power to change your username as well as edit or delete past comments.
    You may opt-in to get notified when people reply.
    </p>

    <p>
    Talk to you soon!
    </p>
  </body>
</html>
"""

WELCOME_2_TEXT = """
Hello again!

Here is the verification code you requested:

 \n{0}\n

Type code into the challenge input field.

Don't forget to check out your notification settings.

Talk to you soon!
"""

WELCOME_2_HTML = """
<!DOCTYPE html>
<html>
<head>
<title>{0}</title>
</head>
  <body>
    <h2>Hello again!</h2>

    <p>
    Here is the verification code you requested:
    </p>

    <p><b>{1}</b></p>

    <h3>What's next?</h3>

    <p>
    Don't forget to check out your notification settings.
    </p>

    <p>
    Talk to you soon!
    </p>
  </body>
</html>
"""

OPERATOR_HTML = """<!DOCTYPE html>
<html>
<head>
<title>Operator Notification</title>
</head>
  <body>
    <h2>Operator Notification!</h2>
    <p>{}</p>
  </body>
</html>
"""



"""locked-rb-node

Revision ID: a38c12eeff2d
Revises: 76deff88cfd6
Create Date: 2020-04-20 10:22:43.611168

"""

# revision identifiers, used by Alembic.
revision = 'a38c12eeff2d'
down_revision = '76deff88cfd6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('rb_node', sa.Column('locked', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('rb_node', 'locked')



"""Added email_id column to User table.

Revision ID: 6a440d3921d2
Revises: 5a5d0b68db78
Create Date: 2017-05-16 16:22:51.161059

"""

# revision identifiers, used by Alembic.
revision = "6a440d3921d2"
down_revision = "5a5d0b68db78"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import User
from remarkbox.models import now_timestamp


def upgrade():

    bind = op.get_bind()
    DBSession = sa.orm.Session(bind=bind)

    op.add_column("User", sa.Column("email_id", sa.Unicode))
    op.add_column(
        "User",
        sa.Column(
            "password_timestamp", sa.BigInteger, server_default=str(now_timestamp())
        ),
    )

    users = DBSession.query(User).all()

    for user in users:
        user.email_id = user._generate_raw_password(8)
        DBSession.add(user)

    DBSession.flush()
    DBSession.commit()


def downgrade():
    # this will fail for sqlite3 databases...
    op.drop_column("User", "email_id")
    op.drop_column("User", "password_timestamp")



"""Namespace stylesheet_embed column

Revision ID: ba8e83a3b8fd
Revises: 87fbd2b4e06e
Create Date: 2017-10-01 11:00:26.447925

"""

# revision identifiers, used by Alembic.
revision = "ba8e83a3b8fd"
down_revision = "87fbd2b4e06e"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "Namespace", sa.Column("stylesheet_embed", sa.UnicodeText(), nullable=True)
    )
    op.add_column(
        "Namespace",
        sa.Column("stylesheet_embed_timestamp", sa.BigInteger(), nullable=True),
    )


def downgrade():
    op.drop_column("Namespace", "stylesheet_embed_timestamp")
    op.drop_column("Namespace", "stylesheet_embed")



"""Add created_timestamp column to Oauth class, add email_notify enum to Namespace

Revision ID: a26302bcfa25
Revises: 00211ed642cc
Create Date: 2018-05-26 19:18:53.858014

"""

# revision identifiers, used by Alembic.
revision = "a26302bcfa25"
down_revision = "00211ed642cc"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_oauth",
        sa.Column(
            "created_timestamp",
            sa.BigInteger(),
            nullable=False,
            server_default="1527378962438",
        ),
    )
    op.add_column(
        "rb_namespace",
        sa.Column(
            "email_notify",
            sa.Enum("immediately", "daily", "weekly", name="email_notify"),
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column("rb_oauth", "created_timestamp")
    op.drop_column("rb_namespace", "email_notify")



from sqlalchemy import BigInteger, Boolean, Column, UnicodeText

from sqlalchemy.orm import relationship, backref

import uuid

from .meta import Base, RBase
from .meta import UUIDType
from .meta import now_timestamp, foreign_key

import json

from .node import get_node_count_who_share_root, get_visible_node_count_who_share_root


class NodeCache(RBase, Base):
    """
    The NodeCache class duplicates and denormalizes results of intensive
    queries, like full table scans, into a lookup table indexed by the node_id.

    You can think of this table and class as a cache for statistics.

    Always invalidate before writing to the underlying normalized relations.
    """

    id = Column(UUIDType, primary_key=True, index=True)
    node_id = Column(UUIDType, foreign_key("Node", "id"), index=True, default=None)
    valid = Column(Boolean, default=False)
    updated_timestamp = Column(BigInteger, nullable=False)
    # Statistics stored as JSON in the database.
    json_stats = Column(UnicodeText, default=None)

    # lazy='joined' performs a left join and reduces amount of queries
    #  from N different users to 1. It's magic.
    node = relationship(
        argument="Node",
        lazy="joined",
        back_populates="cache",
        uselist=False,
        foreign_keys=[node_id],
    )

    def __init__(self, node_id):
        self.id = uuid.uuid1()
        self.node_id = node_id
        self.updated_timestamp = now_timestamp()

    def _recompute_stats(self, new_stats=None):
        # todo: this might make recomputations automatic / easier.
        # https://sqlalchemy-utils.readthedocs.io/en/latest/aggregates.html
        if new_stats is None:
            new_stats = {}
            if self.node.is_root:
                new_stats["root"] = {}
                # don't consider the root node itself in the counts, so minus 1.
                new_stats["root"]["count"] = (
                    get_node_count_who_share_root(self.dbsession, self.node) - 1
                )
                new_stats["root"]["visible_count"] = (
                    get_visible_node_count_who_share_root(self.dbsession, self.node) - 1
                )

        self.valid = True
        self.json_stats = json.dumps(new_stats)
        self.updated_timestamp = now_timestamp()

    def invalidate(self):
        """invalidate stats to force a full recomputation on next read request."""
        self.valid = False
        self.updated_timestamp = now_timestamp()

    @property
    def stats(self):
        if not self.valid:
            self._recompute_stats()
        return json.loads(self.json_stats)

    @stats.setter
    def stats(self, new_stats):
        self._recompute_stats(new_stats)


def invalidate_all_node_cache_objects(dbsession):
    """Only use this if you change the NodeCache stats JSON schema."""
    dbsession.query(NodeCache).update(
        {"valid": False, "updated_timestamp": now_timestamp()}
    )
    dbsession.flush()
